/* 
 * Copyright (C) 2004  Free Software Initiative of Japan
 *
 * This file is subject to the terms and conditions of the GNU Lesser
 * General Public License version 2.1 (or later).
 *
 */

#include "io.h"
#include "led.h"
#include "delay.h"
#include "char-io.h"
#include "conf-ata.h"

static int
cfc_init (void)			/* CF controller initialization */
{
  outw ((inw (CPCR) | 0x0001), CPCR);

  outw (0x0001, CFBUFCR);	/* Disable CF Buffer */
  outw (0x0101, CFRSTCR);	/* CF reset=on, assert OE */
  mdelay (10);

  /* Initial condition: CF insertion interrupt disabled */
  if (inw (CFSTS) == 1)
    {
      mdelay (50);
      outw (0x0000, CFBUFCR);	/* Enable CF Buffer */
      mdelay (10);
		
      outw (0x0100, CFRSTCR);	/* CF reset=off, assert OE */
      mdelay (50);

      return 0;
    }

  outw (0x0001, CFBUFCR);	/* Disable CF Buffer */
  return -1;
}

static void
card_init (void)		/* Card initialization */
{
  outb (COR_SRESET, CARD_COR);
  mdelay (100);
  outb (0x00, CARD_COR);
  mdelay (1);

  outb (0x00, CARD_SCR);
  mdelay (1);
  outb (0x00, CARD_CSR);
  mdelay (1);

  outb (COR_LEVEL_IRQ | COR_IO_MAP, CARD_COR);
  mdelay (1);

  while ((inb (CARD_PRR) & PRR_READY) == 0)
    mdelay (1);
}

#define	IDLE			0xE3
#define	IDENTIFY_DRIVE		0xEC
#define READ_SECTORS		0x20

void
ata_reset (void)
{
  unsigned short w;

  outb (0x06, ATA_CTRL_STAT);	/* Reset */
  mdelay (1);

  outb (0x02, ATA_CTRL_STAT);	/* polling */
  mdelay (10);

  while ((w = inb (ATA_CTRL_STAT)) & 0x80)
    mdelay (50);

  w = inb (ATA_ERROR);
#if defined(DEBUG_ATA)
  put_hex (w);
  put_string ("... ATA Error register (expected '01')\n");
#endif
}

static int
ata_device_select (void)
{
  unsigned short w;

  while ((w = inb (ATA_STAT_CMD)) & 0x88)
    mdelay (1);
#if defined(DEBUG_ATA)
  put_hex (w);
  put_string ("... ATA Status\n");
#endif

  outb (0x02, ATA_CTRL_STAT);	/* polling */
  outb (0x00, ATA_DRV_HD);
  mdelay (1);

  while ((w = inb (ATA_STAT_CMD)) & 0x88)
    mdelay (1);
#if defined(DEBUG_ATA)
  put_hex (w);
  put_string ("... ATA Status (Ready)\n");
#endif

  return 0;
}

static int
ata_register_check (void)
{
  outw (0x55aa, ATA_CYLINDER);
  if (inw (ATA_CYLINDER) != 0x55aa)
    put_string ("Bad ATA registers.\n");
  else
    put_string ("Good ATA registers.\n");
}

static int
ata_cmd (int cmd)
{
  unsigned short w;

  ata_device_select ();
  outb (cmd, ATA_STAT_CMD);
  mdelay (1);

  while ((w = inb (ATA_CTRL_STAT)) & 0x80)
    mdelay (1);

  return w;
}

int
ata_init (void)
{
  unsigned short w;
  unsigned char buf[512];
  int i;

  if (cfc_init () < 0)
    {
      put_string ("*** CFC init failed.\n");
      return -1;
    }

  put_string ("CFC init done.\n");

  card_init ();
  put_string ("Card init done.\n");

  ata_reset ();

  ata_device_select ();

  ata_register_check ();

  ata_cmd (IDENTIFY_DRIVE);
  for (i=0; i<256; i++)
    {
      w = inw (ATA_DATA);
      buf[i*2] = (w & 0xff);
      buf[i*2+1] = (w >> 8);
    }

  while ((w = inb (ATA_CTRL_STAT)) & 0x80)
    mdelay (1);
  w = inb (ATA_STAT_CMD);
#if defined(DEBUG_ATA)
  put_hex (w);
  put_string ("... ATA Status (Done)\n");
#endif

  put_string ("ATA Identity:\n\t");
  buf[54+40] = 0;
  put_string (buf + 54);
  put_string ("\n\t");
  buf[46+8] = 0;
  put_string (buf + 46);
  put_string ("\n\t");
  buf[20+20] = 0;
  put_string (buf + 20);
  put_string ("\n");

  ata_device_select ();
  outb (0, ATA_SECTOR_COUNT);
  outb (IDLE, ATA_STAT_CMD);
  mdelay (1);

  while ((w = inb (ATA_CTRL_STAT)) & 0x80)
    mdelay (1);

  w = inb (ATA_STAT_CMD);
#if defined(DEBUG_ATA)
  put_hex (w);
  put_string ("... ATA Status (Done)\n");
#endif

  return 0;
}

#define ATA_LBA 0x40
#define ATA_ERR 0x01


int
ata_read_sectors (int unit, unsigned int lba,
		  unsigned char *buf, int count)
{
  unsigned short w;
  int i;

  ata_device_select ();

  outb (count, ATA_SECTOR_COUNT);
  outb (lba & 0xff, ATA_SECTOR_NUMBER);
  outb ((lba & 0xff00)>>8, ATA_CYLINDER_LOW);
  outb ((lba & 0xff0000)>>16, ATA_CYLINDER_HIGH);
  outb (ATA_LBA|(unit?0x10:0)|((lba & 0x0f000000)>>24), ATA_DRV_HD);

  outb (READ_SECTORS, ATA_STAT_CMD);
  mdelay (1);

  while (count--)
    {
      while ((w = inb (ATA_CTRL_STAT)) & 0x80)
	mdelay (1);

      if ((w & ATA_ERR))
	{
	  w = inb (ATA_STAT_CMD);
	  put_hex (w);
	  put_string ("... ATA Error\n");
	  return -1;
	}

      for (i=0; i<256; i++)
	{
	  w = inw (ATA_DATA);
	  *buf++ = (w >> 8);
	  *buf++ = (w & 0xff);
	}
    }

  while ((w = inb (ATA_CTRL_STAT)) & 0x80)
    mdelay (1);

  w = inb (ATA_STAT_CMD);
#if defined(DEBUG_ATA)
  put_hex (w);
  put_string ("... ATA Status (Done)\n");
#endif
  return 0;
}

void
ata_fini (void)
{
  outw (0x0000, CFRSTCR);	/* CF reset=off, dis assert OE */
  outw (0x0001, CFBUFCR);	/* Disable CF Buffer */
}
