/* 
 * Copyright (C) 2004, 2005  Free Software Initiative of Japan
 *
 * Author: NIIBE Yutaka <gniibe@fsij.org>
 *
 * This file is subject to the terms and conditions of the GNU Lesser
 * General Public License version 2.1 (or later).
 *
 */

#include "types.h"
#include "char-io.h"
#include "delay.h"
#include "eth.h"
#include "string.h"
#include "conf-mem.h"

#define TRANSMISSION_TIMEOUT		1000
#define RECEIVE_WAIT_1_TIMEOUT		10
#define RECEIVE_WAIT_2_TIMEOUT		10

#define ETH_RECEIVE_WAIT_1		1
#define ETH_RECEIVE_WAIT_2 		2

#define PACKET_OVERHEAD 6 /* Status (2), Count (2), Control byte, Odd byte */

static inline unsigned int eth_inb(unsigned int port)
{
  return *(unsigned char volatile *)(port+IO_BASE);
}

static inline unsigned int eth_inw(unsigned int port)
{
  unsigned short v = *(unsigned short volatile *)(port+IO_BASE);
  return ((v>>8)|((v&0xff)<<8));
}

static inline void eth_insb(unsigned int port, unsigned char *buf,
			    unsigned long count) 
{
  while (count--) *buf++ = *(unsigned char *volatile)(port+IO_BASE);
}

static inline void eth_insw(unsigned int port, unsigned char *addr,
			    unsigned long count)
{
  unsigned short *buf = (unsigned short *)addr;

  while (count--)
    *buf++ = *(unsigned short volatile *)(port+IO_BASE);
}

static inline void eth_outb(unsigned char b, unsigned int port)
{
  *(unsigned char volatile *)(port+IO_BASE) = b;
}

static inline void eth_outw(unsigned short w, unsigned int port)
{
  *(unsigned short volatile *)(port+IO_BASE) = (w>>8|w<<8);
}

void eth_outsw(unsigned int port, const unsigned char *addr,
	       unsigned long count)
{
  const unsigned short *buf = (const unsigned short *)addr;
  while (count--)
    *(unsigned short volatile *)(port+IO_BASE) = *buf++;
}

struct eth_addr node_eth_addr;

/* Transmit Frame Status */
#define TFS_UNDERRUN	0x8000
#define TFS_EXDEFER	0x0800
#define TFS_CARRIER	0x0400
#define TFS_LATE	0x0200
#define TFS_16COL	0x0010

#define TFS_ERROR (TFS_UNDERRUN|TFS_EXDEFER|TFS_CARRIER|TFS_LATE|TFS_16COL)

/* Receive Frame Status */
#define RFS_ALIGN	0x8000
#define RFS_CRC		0x2000
#define RFS_ODD		0x1000
#define RFS_LONG	0x0800
#define RFS_SHORT	0x0400

#define RFS_ERROR (RFS_ALIGN|RFS_CRC|RFS_LONG|RFS_SHORT)

#define BANK_REG			14

/* Bank 0 registers */
#define TX_CTRL_REG			0
#    define TCR_TX_ENABLE	0x0001
#    define TCR_PAD_ENABLE	0x0080
#    define TCR_MON_CSN		0x0400
#    define TCR_FULL_DUPLEX	0x0800
#    define TCR_STOP_SQET	0x1000
#    define TCR_INIT	(TCR_TX_ENABLE|TCR_PAD_ENABLE|TCR_MON_CSN|TCR_FULL_DUPLEX|TCR_STOP_SQET)
#define RX_CTRL_REG			4
#    define RC_RESET		0x8000
#    define RC_RX_ENABLE	0x0100
#    define RC_ABORT		0x0001
#define RX_PHY_CTRL_REG		10
#    define RPCR_ANEG		0x0800

/* Bank 1 registers */
#define CONFIG_REG			0
#    define CFG_EPH_POW_ENABLE	0x8000
#define INDIVIDUAL_ADDRESS_REG		4
#define CTRL_REG			12
#    define	TX_ERR_ENABLE	0x0020		

/* Bank 2 registers */
#define MMU_CMD_REG			0
#    define MMU_BUSY		0x0001
#    define MMU_CMD_ALLOC	0x0020
#    define MMU_CMD_RM_RELEASE	0x0080
#    define MMU_CMD_RELEASE	0x00a0
#    define MMU_CMD_TX_ENQUEUE	0x00c0
#define PACKET_NR_REG			2
#define ALLOC_RES_REG			3
#    define AR_FAILED		0x80
#define TX_FIFO_REG			4
#    define TX_FIFO_EMPTY	0x80
#define RX_FIFO_REG			5
#    define RX_FIFO_EMPTY	0x80
#define POINTER_REG			6
#    define PTR_RECEIVE		0x8000
#    define PTR_AUTOINC		0x4000
#    define PTR_READ		0x2000
#    define PTR_NOT_EMPTY	0x0800
#define DATA_REG			8
#define INTR_REG			12
#    define INTR_RX_DONE	0x01
#    define INTR_TX_EMPTY	0x02
#    define INTR_ALLOC		0x08
#    define INTR_RX_OVERRUN	0x10
#    define INTR_EPH		0x20
#define INTR_MASK_REG			13

/* Bank 3 registers */
#define MT01				0
#define MT23				2
#define MT45				4
#define MT67				6
#define MGMT				8

/*
 * @BANK 0..3
 */
static inline void select_bank (int bank)
{
  eth_outw (bank, BANK_REG);
}

static void mii_zzzz (void)
{
  eth_outw (0x3330, MGMT);		/* MDOE=0 MCLK=0 MDO=0 */
  udelay(40);
  eth_outw (0x3334, MGMT);		/* MDOE=0 MCLK=1 MDO=0 */
  udelay(40);
}

static void mii_zero (void)
{
  eth_outw(0x3338, MGMT);		/* MDOE=1 MCLK=0 MDO=0 */
  udelay(40);
  eth_outw(0x333c, MGMT);		/* MDOE=1 MCLK=1 MDO=0 */
  udelay(40);
}

static void mii_one (void)
{
  eth_outw(0x3339, MGMT);		/* MDOE=1 MCLK=0 MDO=1 */
  udelay(40);
  eth_outw(0x333d, MGMT);		/* MDOE=1 MCLK=1 MDO=1 */
  udelay(40);
}

static void mii_init (void)
{
  mdelay (1000);

  select_bank (0);
  eth_outw (RPCR_ANEG, RX_PHY_CTRL_REG);

  select_bank (3);
  eth_outw(0x3339, MGMT);		/* MDOE=1 MCLK=0 MDO=1 */
  udelay(40);

  /* Synchronize... */
  mii_one ();   mii_one ();  mii_one ();   mii_one ();
  mii_one ();   mii_one ();  mii_one ();   mii_one ();
  mii_one ();   mii_one ();  mii_one ();   mii_one ();
  mii_one ();   mii_one ();  mii_one ();   mii_one ();
  mii_one ();   mii_one ();  mii_one ();   mii_one ();
  mii_one ();   mii_one ();  mii_one ();   mii_one ();
  mii_one ();   mii_one ();  mii_one ();   mii_one ();
  mii_one ();   mii_one ();  mii_one ();   mii_one ();

  mii_zero ();  mii_one ();	/* Start code */
  mii_zero ();  mii_one ();	/* READ=0 WRITE=1 */

  /* PHY Address = 00000 */
  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();

  /* REG Address = 00000 */
  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();

  /* Turn around */
  mii_one ();   mii_zero ();  

  /* All zero for now (XXX Auto negotiation enable/reset doesn't work...) */
  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();
  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();
  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();
  mii_zero ();  mii_zero ();  mii_zero ();  mii_zero ();

  mii_zzzz ();

  eth_outw(0x3330, MGMT);		/* MDOE=0 MCLK=0 MDO=0 */
  mdelay (1000);
}

static void
eth_reset (void)
{
}

/**
 * eth_init - Initialize ethernet packet (which include state information)
 * @p: pointer to ethernet frame and state
 */
/*
 * This software only!
 *
 * Allocate transmit buffer at initialization time, reuse this single
 * buffer for every transmit.
 */
int
eth_init (struct eth_packet *p)
{
  int i;
  unsigned char packet_nr;

  p->state = ETH_IDLE;
  p->data_to_send = 0;
  p->tx_completion_wait = 0;
  p->wait = 0;

  if ((eth_inw (BANK_REG) & 0xff00) != 0x3300)
    return -1;

  select_bank (2);
  eth_outb (0, INTR_MASK_REG);		/* Disable interrupt */

  select_bank (0);
  eth_outw (RC_RESET, RX_CTRL_REG);
  udelay(10);
  eth_outw (0, RX_CTRL_REG);

  mii_init ();

  select_bank (1);
  eth_outw (TX_ERR_ENABLE, CTRL_REG);

  select_bank (0);
  eth_outw (TCR_INIT, TX_CTRL_REG);

  select_bank (1);
  for (i=0; i<ETH_ALEN; i++) 
    {
      node_eth_addr.bits[i] = eth_inb (INDIVIDUAL_ADDRESS_REG+i);
      udelay (10);
    }

  put_string ("eth: MAC Address:");
  for (i=0; i<ETH_ALEN; i++) 
    put_hex (node_eth_addr.bits[i]);

  select_bank (3);
  eth_outw (0, MT01);
  eth_outw (0, MT23);
  eth_outw (0, MT45);
  eth_outw (0, MT67);

  select_bank (1);
  eth_outw (eth_inw (CONFIG_REG)|CFG_EPH_POW_ENABLE, CONFIG_REG);

#define ALLOC_WAIT	10
  select_bank (2);
  eth_outw (MMU_CMD_ALLOC|((ETHER_FRAME_LEN+PACKET_OVERHEAD)>>8), MMU_CMD_REG);
  for (i=0; i<ALLOC_WAIT; i++)
    if ((eth_inb (INTR_REG) & INTR_ALLOC) == 0)
      udelay (10);
    else
      {
	eth_outb (INTR_ALLOC, INTR_REG);
	break;
      }

  if (i >= ALLOC_WAIT || ((packet_nr = eth_inb (ALLOC_RES_REG)) & AR_FAILED))
    {
      put_string ("eth: Memory Allocation failed\n");
      return -1;
    }
  eth_outb (packet_nr, PACKET_NR_REG);

  select_bank (2);
  eth_outb (0, INTR_MASK_REG);

  select_bank (0);
  eth_outw (RC_RX_ENABLE, RX_CTRL_REG);

  return 0;
}

#define CRCPOLY_LE 0xedb88320
static unsigned int crc32_le(unsigned char const *p, int len)
{
  int i, j;
  unsigned int crc = ~0;

  for (i = 0; i < len; i++)
    {
      crc ^= *p++;
      for (j = 0; j < 8; j++)
	crc = (crc >> 1) ^ ((crc & 1) ? CRCPOLY_LE : 0);
    }

  return crc;
}

/**
 * eth_mc_setup_ipv6 - Setup ethenet multicast for IPv6 (all node and link solicitation)
 */
int
eth_mc_setup_ipv6 (void)
{
 /* The bit for 'all node' (0x333300000001) is pre-set. */
  unsigned char multicast_table[8] = {0, 0, 0, 0, 0, 0, 0, 0x40};
  static const unsigned char invert3[] = {0, 4, 2, 6, 1, 5, 3, 7};
  int pos;
  unsigned char mc_link[6] = {0x33, 0x33, 0xff, 0x00, 0x00, 0x00};

  memcpy (&mc_link[3], &node_eth_addr.bits[3], 3);
  pos = crc32_le(mc_link, ETH_ALEN) & 0x3f;
  multicast_table[invert3[pos&7]] |= (1<<invert3[(pos>>3)&7]);

  select_bank (3);
  eth_outw (multicast_table[0] | (multicast_table[1] << 8), MT01);
  eth_outw (multicast_table[2] | (multicast_table[3] << 8), MT23);
  eth_outw (multicast_table[4] | (multicast_table[5] << 8), MT45);
  eth_outw (multicast_table[6] | (multicast_table[7] << 8), MT67);
  select_bank (2);
  return 0;
}

static int
eth_receive (struct eth_packet *p)
{
  unsigned int r;

  select_bank (2);
  if (((r = eth_inb (INTR_REG))
       & (INTR_RX_DONE|INTR_RX_OVERRUN|INTR_EPH)) == 0)
    return -1;

  if ((r & INTR_RX_DONE))
    {
      if ((eth_inb (RX_FIFO_REG) & RX_FIFO_EMPTY) == 0)
	{
	  eth_outw (PTR_AUTOINC|PTR_RECEIVE|PTR_READ, POINTER_REG);
	  return 0;
	}

      /* No packet received. */
      eth_outb (INTR_RX_DONE, INTR_REG);
    }

  if ((r & INTR_RX_OVERRUN))
    {
      select_bank (0);
      eth_outw (eth_inw (RX_CTRL_REG) & ~RC_ABORT, RX_CTRL_REG);

      select_bank (2);
      eth_outb (INTR_RX_OVERRUN, INTR_REG);
    }

  if ((r & INTR_EPH))
    {				/* Re-enable TX */
      select_bank (0);
      eth_outw (eth_inw (TX_CTRL_REG)|TCR_TX_ENABLE, TX_CTRL_REG);
    }

  return -1;
}

static int
eth_receive_wait_1 (struct eth_packet *p)
{
  unsigned short status;
  unsigned int len;

  select_bank (2);
  if ((eth_inw (POINTER_REG) & PTR_NOT_EMPTY) != 0)
    return -1;

  status = eth_inw (DATA_REG);
  if ((status & RFS_ERROR))
    {
      put_string ("eth: Receive error:");
      put_hex (status>>8);
      put_hex (status);
      put_string ("\n");
    }
  else
    {
      len = ((eth_inw (DATA_REG)&0x07ff) - PACKET_OVERHEAD - 4 /* 4-byte for FCS */
	     + ((status & RFS_ODD)?1:0));
      if (len > ETHER_FRAME_LEN)
	{
	  put_string ("eth: Packet too long\n");
	  len = ETHER_FRAME_LEN;
	}

      eth_insw (DATA_REG, p->data, len>>1);
      if ((len & 1))
	p->data[len - 1] = eth_inb (DATA_REG);
      p->length = len;
    }

  eth_outw (MMU_CMD_RM_RELEASE, MMU_CMD_REG);

  return 0;
}

static int
eth_receive_wait_2 (struct eth_packet *p)
{
  select_bank (2);
  if ((eth_inw (MMU_CMD_REG) & MMU_BUSY) != 0)
    return -1;

  eth_outb (INTR_RX_DONE, INTR_REG);
  return 0;
}

static int
eth_transmit_wait_completion (void)
{
  int result = 0;
  unsigned int r;

  select_bank (2);
  if ((eth_inb (INTR_REG) & INTR_TX_EMPTY) == 0)
    return -1;

  if (((r = eth_inb (TX_FIFO_REG)) & TX_FIFO_EMPTY) != 0)
    return -1;

  eth_outb (r, PACKET_NR_REG);
  eth_outw (PTR_AUTOINC|PTR_READ, POINTER_REG);

  if (((r = eth_inw (DATA_REG)) & TFS_ERROR) != 0)
    {
      put_string ("eth: TX Failed:");
      put_hex (r >> 8);
      put_hex (r);
      put_string ("\n");
      result = -1;
    }

  eth_outb (INTR_TX_EMPTY, INTR_REG);
  return result;
}

void
transmit_prepare (struct eth_packet *p)
{
  if (p->data_to_send)
    put_string ("eth: eth_transmit assertion failure.\n"); /* Go anyway. */

  p->data_to_send = 1;
}

static void
eth_transmit (const struct eth_packet *p)
{
  unsigned int r;

  select_bank (2);
  eth_outw (PTR_AUTOINC, POINTER_REG);

  eth_outw (0, DATA_REG);		/* Space for status word */
  eth_outw (p->length+PACKET_OVERHEAD, DATA_REG);
  eth_outsw (DATA_REG, p->data, (p->length)>>1);

  if ((p->length & 1) == 0)
    eth_outw (0, DATA_REG);
  else
    eth_outw (0x2000 | p->data[p->length - 1], DATA_REG);

  eth_outw (MMU_CMD_TX_ENQUEUE, MMU_CMD_REG);

  select_bank (0);
  if (((r = eth_inw (TX_CTRL_REG)) & TCR_TX_ENABLE) == 0)
    eth_outw (r|TCR_TX_ENABLE, TX_CTRL_REG);
}

int
eth_maton (struct eth_packet *p)
{
  int r = -1;

  /*
   * Transmission has higher priority.
   * (No receive when sending or waiting transmission completion)
   * Lower layer's transmission has higher priority.
   * (Higher layer cannot transmit a packet while lower layer has one already.)
   */
  if (p->state == ETH_IDLE)
    {
      if (p->tx_completion_wait)
	{
	  if (eth_transmit_wait_completion () == 0) /* Done */
	    {
	      p->tx_completion_wait = 0;
	      r = 1;
	    }
	  else
	    if (--p->wait == 0)
	      {
		put_string ("Transmit failure.\n");
		eth_reset ();
		p->tx_completion_wait = 0;
		r = 1;
	      }
	}
      else if (p->data_to_send)
	{
	  eth_transmit (p);
	  p->data_to_send = 0;
	  p->tx_completion_wait = 1;
	  p->wait = TRANSMISSION_TIMEOUT;
	  r = 1;
	}
      else
	{
	  if (eth_receive (p) == 0)
	    {
	      p->state = ETH_RECEIVE_WAIT_1;
	      p->wait = RECEIVE_WAIT_1_TIMEOUT;
	      r = 1;
	    }
	  else
	    r = -1;
	}
    }
  else if (p->state == ETH_RECEIVE_WAIT_1)
    {
      if (eth_receive_wait_1 (p) == 0)
	{
	  p->state = ETH_RECEIVE_WAIT_2;
	  p->wait = RECEIVE_WAIT_2_TIMEOUT;
	  r = 0;
	}
      else
	if (--p->wait == 0)
	  {
	    put_string ("eth: Receive failure.\n");
	    eth_reset ();
	    eth_outb (INTR_RX_DONE, INTR_REG);
	    p->state = ETH_IDLE;
	  }
    }
  else if (p->state == ETH_RECEIVE_WAIT_2)
    {	     /* We cannot invoke eth_transmit while waiting MMU_CMD_REG */
      if (eth_receive_wait_2 (p) == 0)
	{
	  p->state = ETH_IDLE;
	  r = 1;
	}
      else
	if (--p->wait == 0)
	  {
	    put_string ("eth: Receive failure (MMU_CMD busy).\n");
	    eth_reset ();
	    eth_outb (INTR_RX_DONE, INTR_REG);
	    p->state = ETH_IDLE;
	  }
    }

  return r;
}
