/* 
 * Copyright (C) 2004  Free Software Initiative of Japan
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License version 2 (or later).
 *
 */

#include "types.h"
#include "string.h"
#include "pp.h"
#include "eth.h"
#include "delay.h"
#include "char-io.h"
#include "netinet.h"

/*
 * IP:		node_ip_addr
 * Netmask:	node_netmask
 */

u32 node_ip_addr;
u32 node_netmask;

void net_interface_config (u32 ip, u32 netmask)
{
  node_ip_addr = ip;
  node_netmask = netmask;
}

int inet_aton (const char *str, u32 *ip_addr_p)
{
  u32 ip=0;
  u8 byte;
  u8 c;
  int i;

  for (i=0; i<4; i++)
    {
      byte = 0;
      while ((c = *str++) && c >= '0' && c <= '9')
	byte = byte * 10 + c - '0';
      ip = (ip << 8) + byte;
      if (i < 3 && c != '.')
	return -1;
    }

  *ip_addr_p = ip;
  return 0;
}

/* Assume 2-byte alignment */
static u16 in_checksum_sub (void *p, int len, u16 initial)
{
  u16 *wp = (u16 *)p;
  u32 sum = initial;

  while (len >= 2)
    {
      sum += *wp++;
      len -= 2;
    }

  if (len)
    {
      union { u8 c[2]; u16 w; } x;

      x.c[0] =  *(u8 *)wp;
      x.c[1] = 0;
      sum += x.w;
    }

  {
    union { u16 w[2]; u32 l; } x;

    x.l = sum;
    sum = x.w[0] + x.w[1];
    if (sum > 0xffff)
      sum -= 0xffff;
  }

  return (~sum & 0xffff);
}

static u16 in_checksum (void *p, int len)
{
  return in_checksum_sub (p, len, 0);
}

static void ether_header_build (struct eth_packet *p,
				const struct eth_addr *dst_eth_addr, u16 type)
{
  struct eth_frame *e = (struct eth_frame *)p->data;

  memcpy (e->dst, dst_eth_addr->bits, ETH_ALEN);
  memcpy (e->src, node_eth_addr.bits, ETH_ALEN);
  poke_u16_hton (e->type, type);
}

struct arp {
  u8 hw_type[2];
  u8 proto_type[2];
  u8 hw_size[1];
  u8 proto_size[1];
  u8 op[2];
  u8 sender_eth_addr[ETH_ALEN];
  u8 sender_ip_addr[4];
  u8 target_eth_addr[ETH_ALEN];
  u8 target_ip_addr[4];
};

#define ARP_REQUEST	1
#define ARP_REPLY	2

const struct eth_addr eth_broadcast = {{0xff, 0xff, 0xff, 0xff, 0xff, 0xff}};

/*
 * This software only!
 *
 * We never timeout arp cache.
 * We only support Ethernet.
 * We only support IPv4.
 */

/*
 * mk_arp_request
 * 
 * @ip: IP address to resolve.
 * @p: packet.
 */
static void mk_arp_request (struct eth_packet *p, u32 ip)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct arp *a = (struct arp *)e->data;

  p->length = ETH_HLEN + sizeof (struct arp);
  ether_header_build (p, &eth_broadcast, ETHERTYPE_ARP);
  poke_u16_hton (a->hw_type, 1);
  poke_u16_hton (a->proto_type, ETHERTYPE_IP);
  poke_u8 (a->hw_size, 6);
  poke_u8 (a->proto_size, 4);
  poke_u16_hton (a->op, ARP_REQUEST);
  memcpy (a->sender_eth_addr, node_eth_addr.bits, ETH_ALEN);
  poke_u32_hton (a->sender_ip_addr, node_ip_addr);
  memset (a->target_eth_addr, 0, ETH_ALEN);
  poke_u32_hton (a->target_ip_addr, ip);
}

static int arp_request_serv (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct arp *a = (struct arp *)e->data;

  if (peek_u16_ntoh (a->hw_type) != 1)
    return 0;

  if (peek_u16_ntoh (a->proto_type) != ETHERTYPE_IP)
    return 0;

  if (peek_u32_ntoh (a->target_ip_addr) != node_ip_addr)
    return 0;

  /* XXX: More sanity check... */

  p->length = ETH_HLEN + sizeof (struct arp);
  ether_header_build (p, (struct eth_addr *)e->src, ETHERTYPE_ARP);
  poke_u16_hton (a->op, ARP_REPLY);
  memcpy (a->target_eth_addr, a->sender_eth_addr, ETH_ALEN);
  memcpy (a->target_ip_addr, a->sender_ip_addr, 4);
  memcpy (a->sender_eth_addr, node_eth_addr.bits, ETH_ALEN);
  poke_u32_hton (a->sender_ip_addr, node_ip_addr);

  transmit_prepare (p);
  return 1;
}

static int handle_arp_reply (struct eth_packet *p, u32 ip_addr,
			     struct eth_addr *eth_addr)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct arp *a = (struct arp *)e->data;

  if (peek_u16_ntoh (e->type) != ETHERTYPE_ARP)
    return 0;

  if (p->length < ETH_HLEN + sizeof (struct arp))
    return 0;

  if (peek_u16_ntoh (a->op) != ARP_REPLY)
    return 0;

  if (peek_u16_ntoh (a->hw_type) != 1)
    return 0;

  if (peek_u16_ntoh (a->proto_type) != ETHERTYPE_IP)
    return 0;

  if (peek_u32_ntoh (a->target_ip_addr) != node_ip_addr)
    return 0;

  if (peek_u32_ntoh (a->sender_ip_addr) != ip_addr)
    return 0;

  /* XXX: More sanity check??? */
  memcpy (eth_addr, a->sender_eth_addr, ETH_ALEN);
  return 1;
}

static int arp_maton (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct arp *a = (struct arp *)e->data;
  int r;

  if ((r = eth_maton (p)) != 0)
    return r;

  if (peek_u16_ntoh (e->type) != ETHERTYPE_ARP)
    return 0;

  if (p->length < ETH_HLEN + sizeof (struct arp))
    /* We can't handle it */
    return 0;

  if (peek_u16_ntoh (a->op) != ARP_REQUEST)
    /* It's ARP reply, not request */
    return 0;

  r = arp_request_serv (p);
  return r;
}

u16 id=0;
#define PROTO_ICMP	 1
#define PROTO_TCP	 6
#define PROTO_UDP	17

#define IP_HEADER_VERSION	((4<<4)|(20>>2))
#define IP_DF			0x4000		/* Don't fragment flag */
#define IP_TTL		64

struct ip {
  u8 version_hlen[1];
  u8 tos[1];
  u8 total_len[2];
  u8 id[2];
  u8 fragmentation[2];
  u8 ttl[1];
  u8 proto[1];
  u8 checksum[2];
  u8 src[4];
  u8 dst[4];
  u8 data[0];
};

#define ICMP_ECHO_REPLY			0
#define ICMP_DESTINATION_UNREACHABLE	3
#define ICMP_SOURCE_QUENCH		4
#define ICMP_REDIRECT			5
#define ICMP_ECHO_REQUEST		8
#define ICMP_PARAMETER_PROBLEM		12
struct icmp {
  u8 type[1];
  u8 code[1];
  u8 checksum[2];
  u8 data[0];
};

struct icmp_echo {
  u8 id[2];
  u8 seq[2];
  u8 data[0];
};

#define IP_TOS_NULL		0x00
#define IP_TOS_MINIMIZE_DELAY	0x10

static void ip_header_build (struct eth_packet *p, u8 proto,
			     u8 tos, u16 total_len)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;

  poke_u8 (ip->version_hlen, IP_HEADER_VERSION);
  poke_u8 (ip->tos, tos);
  poke_u16_hton (ip->total_len, total_len);
  poke_u16_hton (ip->id, id++);
  poke_u16_hton (ip->fragmentation, IP_DF);
  poke_u8 (ip->ttl, IP_TTL);
  poke_u8 (ip->proto, proto);
  poke_u16_raw (ip->checksum, 0);
  poke_u32_hton (ip->src, node_ip_addr);
  poke_u16_raw (ip->checksum, in_checksum (ip, sizeof (struct ip)));
}

static void mk_icmp_echo (struct eth_packet *p,
			  const struct eth_addr *dst_eth_addr, u32 dst_ip_addr,
			  unsigned int icmp_echo_id, unsigned int seq,
			  unsigned char *data, int len)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct icmp *icmp = (struct icmp *)ip->data;
  struct icmp_echo *icmp_echo = (struct icmp_echo *)icmp->data;

  p->length = ETH_HLEN + sizeof (struct ip) + sizeof (struct icmp) + sizeof (struct icmp_echo) + len;
  ether_header_build (p, dst_eth_addr, ETHERTYPE_IP);
  poke_u32_hton (ip->dst, dst_ip_addr);
  ip_header_build (p, PROTO_ICMP, IP_TOS_NULL,
		   sizeof (struct ip) + sizeof (struct icmp)
		   + sizeof (struct icmp_echo) + len);
  poke_u8 (icmp->type, ICMP_ECHO_REQUEST);
  poke_u8 (icmp->code, 0);
  poke_u16_raw (icmp->checksum, 0);
  poke_u16_hton (icmp_echo->id, icmp_echo_id);
  poke_u16_hton (icmp_echo->seq, seq);
  memcpy (icmp_echo->data, data, len);
  poke_u16_raw (icmp->checksum, in_checksum (icmp, sizeof (struct icmp)+sizeof(struct icmp_echo)+len));
}

static int handle_icmp_echo_reply (struct eth_packet *p, u32 ip_addr, int seq)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct icmp *icmp = (struct icmp *)ip->data;
  struct icmp_echo *icmp_echo = (struct icmp_echo *)icmp->data;

  if (peek_u16_ntoh (e->type) != ETHERTYPE_IP)
    return 0;

  if (peek_u8 (ip->proto) != PROTO_ICMP)
    return 0;

  if (p->length < ETH_HLEN + sizeof (struct ip) + sizeof (struct icmp))
    return 0;

  if (peek_u8 (icmp->type) != ICMP_ECHO_REPLY)
    return 0;

  if (p->length < ETH_HLEN + sizeof (struct ip) + sizeof (struct icmp)
      + sizeof (struct icmp_echo)
      || p->length < ETH_HLEN + peek_u16_ntoh (ip->total_len)
      || in_checksum (icmp, peek_u16_ntoh (ip->total_len) - sizeof (struct ip)) != 0)
    return 0;

  if (peek_u16_ntoh (icmp_echo->seq) != seq)
    return 0;

  return 1;
}

static int icmp_echo_request_serv (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct icmp *icmp = (struct icmp *)ip->data;
  int icmplen;

  if (p->length < ETH_HLEN + sizeof (struct ip) + sizeof (struct icmp)
      + sizeof (struct icmp_echo)
      || p->length < ETH_HLEN + peek_u16_ntoh (ip->total_len)
      || in_checksum (icmp, peek_u16_ntoh (ip->total_len) - sizeof (struct ip)) != 0)
    return -1;			/* Discard */

  icmplen = peek_u16_ntoh (ip->total_len) - sizeof (struct ip);

  p->length = ETH_HLEN + sizeof (struct ip) + icmplen;
  ether_header_build (p, (struct eth_addr *)e->src, ETHERTYPE_IP);
  poke_u32_raw (ip->dst, peek_u32_raw (ip->src));
  poke_u32_hton (ip->src, node_ip_addr);
  /* We don't touch IP's tos, identification, ttl and checksum... */
  poke_u8 (icmp->type, ICMP_ECHO_REPLY);
  poke_u16_raw (icmp->checksum, 0);
  poke_u16_raw (icmp->checksum, in_checksum (icmp, icmplen));

  transmit_prepare (p);
  return 1;
}

static int ip_maton (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct icmp *icmp = (struct icmp *)ip->data;
  int r;

  if ((r = arp_maton (p)) != 0)
    return r;

  if (peek_u16_ntoh (e->type) != ETHERTYPE_IP)
    return 0;

  if (p->length < ETH_HLEN + sizeof (struct ip)
      || peek_u8 (ip->version_hlen) != IP_HEADER_VERSION
      || (node_ip_addr && peek_u32_ntoh (ip->dst) != node_ip_addr)
      || in_checksum (ip, sizeof (struct ip)) != 0)
    /* Discard */
    return -1;

  if (peek_u8 (ip->proto) != PROTO_ICMP)
    return 0;

  if (p->length < ETH_HLEN + sizeof (struct ip) + sizeof (struct icmp))
    /* Discard */
    return -1;

  if (peek_u8 (icmp->type) != ICMP_ECHO_REQUEST)
    return 0;

  r = icmp_echo_request_serv (p);
  return r;
}

#define MAX_ARP_TRY 3
int arp_resolv (struct eth_packet *p, u32 dst_ip_addr,
		struct eth_addr *dst_eth_addr)
{
  int try;

  for (try=0; try < MAX_ARP_TRY; try++)
    {
      int i, r;

      mk_arp_request (p, dst_ip_addr);
      transmit_prepare (p);

#define ARP_TIMEOUT 500
      i = 0;
      while (1)
	{
	  while ((r = ip_maton (p)) == 1);

	  if (r == 0 && handle_arp_reply (p, dst_ip_addr, dst_eth_addr) != 0)
	    return 1;

	  mdelay (1);
	  i++; 
	  if (i>=ARP_TIMEOUT)
	    break;
	}
    }

  return -1;
}

int ping (struct eth_packet *p, const struct eth_addr *dst_eth_addr,
	  u32 dst_ip_addr, int count)
{
  int seq = 0;

  while (seq < count)
    {
      int i, r;

      mk_icmp_echo (p, dst_eth_addr, dst_ip_addr, 0xbeef, seq,
		    "Hello TCP/IP! Hello TCP/IP!! Hello TCP/IP!!!", 44);
      transmit_prepare (p);

#define PING_TIMEOUT 1000
      i = 0;
      while (1)
	{
	  while ((r = ip_maton (p)) == 1);

	  if (r == 0 && handle_icmp_echo_reply (p, dst_ip_addr, seq) != 0)
	    return 1;

	  mdelay (1);
	  i++;
	  if (i >= PING_TIMEOUT)
	    break;
	}

      seq++;
    }

  return -1;
}

static u16 tcpudp_checksum (struct ip *ip, u16 proto)
{
  int len, orig_len;
  int total_len = peek_u16_ntoh (ip->total_len);

  orig_len = len = total_len - sizeof (struct ip);
  if ((len & 1))
    {
      poke_u8 (&ip->data[len], 0);
      len++;
    }

  return in_checksum_sub (ip->src, 8+len, proto+orig_len);
}

#define TCP_URG		0x20
#define TCP_ACK		0x10
#define TCP_PSH		0x08
#define TCP_RST		0x04
#define TCP_SYN		0x02
#define TCP_FIN		0x01

#define TCP_ISN		0x3e33268f

struct tcp {
  u8 src_port[2];
  u8 dst_port[2];
  u8 seq[4];
  u8 ack[4];
  u8 hlen[1];
  u8 flags[1];
  u8 window_size[2];
  u8 checksum[2];
  u8 urgent[2];
  u8 data[0];
};

static void tcp_header_build (struct eth_packet *p,
			      u32 seq, u32 ack, u8 hlen, u8 flags, u16 w_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct tcp *tcp = (struct tcp *)ip->data;

  poke_u32_hton (tcp->seq, seq);
  poke_u32_hton (tcp->ack, ack);
  poke_u8 (tcp->hlen, hlen<<2);
  poke_u8 (tcp->flags, flags);
  poke_u16_hton (tcp->window_size, w_size);
  poke_u16_raw (tcp->checksum, 0);
  poke_u16_hton (tcp->urgent, 0);
  poke_u16_raw (tcp->checksum, tcpudp_checksum (ip, PROTO_TCP));
}

#define TCP_OPTION_END  0
#define TCP_OPTION_NOP  1
#define TCP_OPTION_MSS	2

static void
mk_tcp_syn (struct eth_packet *p,
	    const struct eth_addr *dst_eth_addr,
	    u32 dst_ip_addr, u16 dst_port,
	    u16 my_port, u16 mss, u32 seq, u16 window_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct tcp *tcp = (struct tcp *)ip->data;

  p->length = ETH_HLEN + sizeof (struct ip) + sizeof (struct tcp)
    + 4; /* Four bytes for TCP mss option */
  ether_header_build (p, dst_eth_addr, ETHERTYPE_IP);
  poke_u32_hton (ip->dst, dst_ip_addr);
  ip_header_build (p, PROTO_TCP, IP_TOS_MINIMIZE_DELAY,
		   sizeof (struct ip) + sizeof (struct tcp) + 4);
  poke_u16_hton (tcp->src_port, my_port);
  poke_u16_hton (tcp->dst_port, dst_port);
  tcp->data[0] = TCP_OPTION_MSS;
  tcp->data[1] = 4;
  poke_u16_hton (&tcp->data[2], mss);
  tcp_header_build (p, seq, 0, 24, TCP_SYN, window_size);
}

static int tcp_syn_ack_serv (struct eth_packet *p, unsigned int t_seq,
			     u8 *send, int send_size, int window_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct tcp *tcp = (struct tcp *)ip->data;
  u16 port;
  u32 seq;
  int len;

  seq = peek_u32_ntoh (tcp->seq);
  len = send_size + sizeof (struct ip) + sizeof (struct tcp);

  if (peek_u16_ntoh (tcp->window_size) < send_size)
    put_string ("eth: sending TCP data too big!\n"); /* XXX should do something */

  p->length = ETH_HLEN + len;
  ether_header_build (p, (struct eth_addr *)e->src, ETHERTYPE_IP);
  poke_u32_raw (ip->dst, peek_u32_raw (ip->src));
  ip_header_build (p, PROTO_TCP, IP_TOS_MINIMIZE_DELAY, len);
  port = peek_u16_raw (tcp->src_port);
  poke_u16_raw (tcp->src_port, peek_u16_raw (tcp->dst_port));
  poke_u16_raw (tcp->dst_port, port);
  memcpy (tcp->data, send, send_size);
  tcp_header_build (p, t_seq, seq+1, 20, TCP_ACK|TCP_PSH|TCP_FIN, window_size);

  transmit_prepare (p);
  return seq + 1; /* SYN consumes 1-byte. */
}

static void tcp_reset (struct eth_packet *p,
		       const struct eth_addr *dst_eth_addr,
		       u32 dst_ip_addr, u16 dst_port,
		       u16 my_port, u32 seq)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct tcp *tcp = (struct tcp *)ip->data;

  p->length = ETH_HLEN + sizeof (struct ip) + sizeof (struct tcp);
  ether_header_build (p, dst_eth_addr, ETHERTYPE_IP);
  poke_u32_hton (ip->dst, dst_ip_addr);
  ip_header_build (p, PROTO_TCP, IP_TOS_MINIMIZE_DELAY,
		   sizeof (struct ip) + sizeof (struct tcp));
  poke_u16_hton (tcp->src_port, my_port);
  poke_u16_hton (tcp->dst_port, dst_port);
  tcp_header_build (p, seq, 0, 20, TCP_RST, 0xffff);

  transmit_prepare (p);
}

static int tcp_ack (struct eth_packet *p,
		    unsigned int t_seq, unsigned int r_seq, u16 window_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct tcp *tcp = (struct tcp *)ip->data;
  u16 port;
  int r = 0;

  p->length = ETH_HLEN + sizeof (struct ip) + sizeof (struct tcp);
  ether_header_build (p, (struct eth_addr *)e->src, ETHERTYPE_IP);
  poke_u32_raw (ip->dst, peek_u32_raw (ip->src));
  ip_header_build (p, PROTO_TCP, IP_TOS_MINIMIZE_DELAY,
		   sizeof (struct ip) + sizeof (struct tcp));
  port = peek_u16_raw (tcp->src_port);
  poke_u16_raw (tcp->src_port, peek_u16_raw (tcp->dst_port));
  poke_u16_raw (tcp->dst_port, port);
  tcp_header_build (p, t_seq, r_seq, 20, TCP_ACK, window_size);

  transmit_prepare (p);
  return r;
}

#define CONNECTION_TIMEOUT	1000
#define CONNECTION_RETRY	5
#define RETRANSMIT_TIMEOUT	100
#define TIMEWAIT_TIMEOUT	1000

#define IP_LOOP_COUNT  50	/* 50 times nealy equals to 1ms  */

#define ANNOUNCED_MSS		(ETHER_FRAME_LEN-ETH_HLEN-40) /* IP=20 TCP=20 bytes long */

#define WINDOW_SIZE	6000

int tcp_maton (struct eth_packet *p,
	       const struct eth_addr *dst_eth_addr,
	       u32 server_ip_addr, u16 server_port, u16 my_port,
	       u8 *send, int send_size,
	       u8 *receive, int receive_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct icmp *icmp = (struct icmp *)ip->data;
  struct tcp *tcp = (struct tcp *)ip->data;
  int r;
  unsigned int count, retransmit_time;
  int retry = 0;
  unsigned int t_seq = TCP_ISN, r_seq, seq;
  u16 mss = 536;
  u8 c;
  u16 w;
  u32 x;
  int i;
  int len, offset;
  int receive_count = 0;
  int window_size = WINDOW_SIZE; /* Because of ethernet layer */

  /* Connect */
  mk_tcp_syn (p, dst_eth_addr, server_ip_addr, server_port,
	      my_port, ANNOUNCED_MSS, t_seq, window_size);
  transmit_prepare (p);

  count = 1;
  retransmit_time = CONNECTION_TIMEOUT;
  while (1)
    {
      if (count && (int)(retransmit_time - count) < 0) /* Timer expired */
	{
	  mk_tcp_syn (p, dst_eth_addr, server_ip_addr, server_port,
		      my_port, ANNOUNCED_MSS, t_seq, window_size);
	  transmit_prepare (p);

	  ++retry;
	  retransmit_time = (CONNECTION_TIMEOUT<<retry) + count;
	  if (retry > CONNECTION_RETRY)
	    /* Really timeout */
	    return -1;
	}

      i = 0;
      while ((r = ip_maton (p)) == 1)
	i++;
      count += i / IP_LOOP_COUNT;

      if (r < 0 || peek_u16_ntoh (e->type) != ETHERTYPE_IP)
	goto next;

      if (peek_u8 (ip->proto) == PROTO_ICMP)
	{			/* Error? */
	  if (peek_u8 (icmp->type) != ICMP_DESTINATION_UNREACHABLE)
	    goto next;

	  /* ICMP_REDIRECT? ICMP_PARAMETER_PROBLEM? */
	  /* XXX: Should check if this is for this TCP connection */
	  return -1;
	}

      if (peek_u8 (ip->proto) != PROTO_TCP ||
	  p->length < ETH_HLEN + peek_u16_ntoh (ip->total_len) ||
	  peek_u16_ntoh (ip->total_len) < sizeof (struct ip) + sizeof (struct tcp) ||
	  tcpudp_checksum (ip, PROTO_TCP) != 0 ||
	  peek_u32_ntoh (ip->src) != server_ip_addr ||
	  peek_u32_ntoh (ip->dst) != node_ip_addr ||
	  peek_u16_ntoh (tcp->src_port) != server_port ||
	  peek_u16_ntoh (tcp->dst_port) != my_port)
	goto next;

      if ((peek_u8 (tcp->flags) & TCP_RST))
	return -1;

      break;

    next:
      mdelay (1);
      if (count)
	count++;
    }

  t_seq++;
 syn_ack:
  if ((peek_u8 (tcp->flags) & (TCP_SYN|TCP_ACK)) != (TCP_SYN|TCP_ACK))
    {
      tcp_reset (p, dst_eth_addr, server_ip_addr, server_port, my_port, t_seq);
      return -1; /* XXX: upper layer have to call eth_maton to send RST...  */
    }

  if (peek_u32_ntoh (tcp->ack) != t_seq)
    {
      tcp_reset (p, dst_eth_addr, server_ip_addr, server_port, my_port, t_seq);
      return -1; /* XXX: upper layer have to call eth_maton to send RST...  */
    }

  if ((c = peek_u8 (tcp->hlen) >> 2) > sizeof (struct tcp))
    {				/* TCP option */
      int i;
      int len = c - sizeof (struct tcp);
      int data_len;

      /* We only handle MSS */
      for (i=0; i<len; i+= data_len)
	if ((c = peek_u8 (&tcp->data[i])) == TCP_OPTION_MSS
	    && peek_u8 (&tcp->data[i+1]) == 4)
	  {
	    mss = peek_u16_ntoh (&tcp->data[2]);
	    break;
	  }
	else if (c == TCP_OPTION_END)
	  break;
	else if (c == TCP_OPTION_NOP)
	  data_len = 1;
	else
	  data_len = peek_u8 (&tcp->data[i+1]);
    }

  if (mss < send_size)
    put_string ("eth: sending TCP data too big!\n"); /* XXX should do something */

  /* Connection established */
  r_seq = tcp_syn_ack_serv (p, t_seq, send, send_size, window_size);
  t_seq += send_size + 1;	/* FIN consumes 1-byte. */

  /* FIN-WAIT-1 */
  count = 0;
  while (1)
    {
      if (count && (int)(retransmit_time - count) < 0) /* Timer expired */
	break;

      i = 0;
      while ((r = ip_maton (p)) == 1)
	i++;
      count += i / IP_LOOP_COUNT;

      if (r < 0)
	{ /* When we go to idle state, reset window size to max. */
	  if (p->state == ETH_IDLE && !p->tx_completion_wait)
	    window_size = WINDOW_SIZE;
	  goto next_1;
	}

      if (peek_u16_ntoh (e->type) != ETHERTYPE_IP)
	goto next_1;		/* Discard */

      if (peek_u8 (ip->proto) == PROTO_ICMP)
	{			/* Error? */
	  if (peek_u8 (icmp->type) != ICMP_DESTINATION_UNREACHABLE)
	    goto next_1;

	  /* ICMP_REDIRECT? ICMP_PARAMETER_PROBLEM? */
	  /* XXX: Should check if this is for this TCP connection */
	  return -1;
	}

      if (peek_u8 (ip->proto) != PROTO_TCP ||
	  p->length < ETH_HLEN + peek_u16_ntoh (ip->total_len) ||
	  peek_u16_ntoh (ip->total_len) < sizeof (struct ip) + sizeof (struct tcp) ||
	  tcpudp_checksum (ip, PROTO_TCP) != 0 ||
	  peek_u32_ntoh (ip->src) != server_ip_addr ||
	  peek_u32_ntoh (ip->dst) != node_ip_addr ||
	  peek_u16_ntoh (tcp->src_port) != server_port ||
	  peek_u16_ntoh (tcp->dst_port) != my_port)
	goto next_1;		/* Discard */

      if (((w = peek_u8 (tcp->flags)) & TCP_RST))
	return -1;

      if ((x = peek_u32_ntoh (tcp->ack)) != t_seq) /* SYN+ACK again */
	{
	  put_string ("Got SYN+ACK again\n");
	  t_seq -= send_size + 1;
	  goto syn_ack;
	}

      /* FIN-WAIT-2 or CLOSED */

      seq = peek_u32_ntoh (tcp->seq);
      len = peek_u16_ntoh (ip->total_len) - sizeof (struct ip) - (peek_u8 (tcp->hlen)>>2);
      offset = (peek_u8 (tcp->hlen)>>2) - sizeof (struct tcp);

      if (seq == r_seq)		/* New data */
	{
	  static int time = 0;
	  if ((++time & 0x1f) == 0x1f)
	    put_string (".");

	  if (receive_size < len)
	    {
	      put_string ("eth: Receive too big!\n");
	      tcp_reset (p, dst_eth_addr, server_ip_addr, server_port, my_port, t_seq);
	      return -1; /* XXX: upper layer have to call eth_maton to send RST...  */
	    }

	  if (len > 0)
	    {
	      memcpy (receive, tcp->data + offset, len);
	      receive += len;
	      receive_size -= len;
	      receive_count += len;
	      r_seq += len;
	      seq += len;
	      if (window_size > len)
		window_size -= len;
	      else
		window_size = 0;
	    }
	}

      if ((peek_u8 (tcp->flags) & TCP_FIN))
	{
	  if (seq == r_seq)
	    r_seq++; /* Fall through (next condition will met) */

	  if ((int)(r_seq - seq) > 0)
	    {			/* Already got FIN, restart timer */
	      count = 1;
	      retransmit_time = TIMEWAIT_TIMEOUT + count;
	    }
	}

      tcp_ack (p, t_seq, r_seq, window_size);
      continue;

    next_1:
      mdelay (1);
      if (count)
	count++;
    }

  return receive_count;
}

struct udp {
  u8 src_port[2];
  u8 dst_port[2];
  u8 len[2];
  u8 checksum[2];
  u8 data[0];
};

static void udp_header_build (struct eth_packet *p,
			      u16 src_port, u16 dst_port, u16 len)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct udp *udp = (struct udp *)ip->data;
  u16 checksum;

  poke_u16_hton (udp->src_port, src_port);
  poke_u16_hton (udp->dst_port, dst_port);
  poke_u16_hton (udp->len, len);
  poke_u16_raw (udp->checksum, 0);
  checksum = tcpudp_checksum (ip, PROTO_UDP);
  if (checksum == 0)
    checksum = ~0;
  poke_u16_raw (udp->checksum, checksum);
}

#define BOOTP_SERVER	67
#define BOOTP_CLIENT	68

#define BOOTP_REQUEST	1
#define BOOTP_REPLY	2

#define IP_BROADCAST	0xFFFFFFFF

#define MAX_DHCP_OPTION_LEN	1232 /* For ethernet (1500-byte) */
#define MAX_SERVER_NAME 64
#define MAX_FILE_NAME 128

struct dhcp {
  u8 op[1];
  u8 htype[1];
  u8 hlen[1];
  u8 hops[1];
  u8 xid[4];
  u8 secs[2];
  u8 flags[2];
  u8 ciaddr[4];
  u8 yiaddr[4];
  u8 siaddr[4];
  u8 giaddr[4];
  u8 chaddr[16];
  u8 sname[MAX_SERVER_NAME];
  u8 file[MAX_FILE_NAME];
  u8 cookie[4];
  u8 options[MAX_DHCP_OPTION_LEN];
};

#define BOOTP_EXT_COOKIE		  0x63825363
#define BOOTP_EXT_PAD			  0

#define BOOTP_EXT_NETMASK		  1
#define BOOTP_EXT_ROUTER		  3
#define BOOTP_EXT_DNS			  6
#define BOOTP_EXT_HOSTNAME		 12
#define BOOTP_EXT_ROOTPATH		 17

#define BOOTP_EXT_BROADCAST		 28
#define BOOTP_EXT_BOOTFILESIZE		 13
#define BOOTP_EXT_NTPSERVER		 42
#define BOOTP_EXT_VENDOR		 43

#define DHCP_REQ_ADDR			 50
#define DHCP_MSG_TYPE			 53
#define DHCP_SRV_ID			 54
#define DHCP_PRM_LIST			 55
#define DHCP_MAX_SIZE			 57
#define BOOTP_EXT_SITESPECIFIC0		128
#define BOOTP_EXT_SITESPECIFIC1		129
#define BOOTP_EXT_END			255

#define DHCPDISCOVER		1
#define DHCPOFFER		2
#define DHCPREQUEST		3
#define DHCPACK			5

#define DHCP_MAX_MSG (sizeof (struct udp) + sizeof (struct dhcp))

static const u8 dhcpdiscover[]={
  DHCP_MSG_TYPE, 1, DHCPDISCOVER,
  DHCP_MAX_SIZE, 2,
  DHCP_MAX_MSG>>8, DHCP_MAX_MSG&0xff,
  DHCP_PRM_LIST, 5,
  BOOTP_EXT_NETMASK, BOOTP_EXT_ROUTER,
  BOOTP_EXT_DNS, BOOTP_EXT_HOSTNAME, BOOTP_EXT_ROOTPATH,
  BOOTP_EXT_END
};

static const u8 dhcprequest[]={
  DHCP_MSG_TYPE, 1, DHCPREQUEST,
  DHCP_SRV_ID, 4, 0, 0, 0, 0,
  DHCP_REQ_ADDR, 4, 0, 0, 0, 0,
  DHCP_MAX_SIZE, 2,
  DHCP_MAX_MSG>>8, DHCP_MAX_MSG&0xff,
  DHCP_PRM_LIST, 5,
  BOOTP_EXT_NETMASK, BOOTP_EXT_ROUTER,
  BOOTP_EXT_DNS, BOOTP_EXT_HOSTNAME, BOOTP_EXT_ROOTPATH,
  BOOTP_EXT_END
};

static void mk_dhcp (struct eth_packet *p, u32 xid, u16 secs,
		     int is_discover, u32 server_id, u32 req_addr)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct udp *udp = (struct udp *)ip->data;
  struct dhcp *dhcp = (struct dhcp *)udp->data;
  int dhcp_size = sizeof (struct dhcp) - MAX_DHCP_OPTION_LEN;

  dhcp_size += is_discover? sizeof dhcpdiscover: sizeof dhcprequest;
  p->length = ETH_HLEN + sizeof (struct ip) + sizeof (struct udp) + dhcp_size;
  ether_header_build (p, &eth_broadcast, ETHERTYPE_IP);
  poke_u32_raw (ip->dst, IP_BROADCAST);
  ip_header_build (p, PROTO_UDP, IP_TOS_MINIMIZE_DELAY,
		   sizeof (struct ip) + sizeof (struct udp) + dhcp_size);
  poke_u8 (dhcp->op, BOOTP_REQUEST);
  poke_u8 (dhcp->htype, 1);
  poke_u8 (dhcp->hlen, ETH_ALEN);
  poke_u8 (dhcp->hops, 0);
  poke_u32_hton (dhcp->xid, xid);
  poke_u16_hton (dhcp->secs, secs);
  poke_u16_raw (dhcp->flags, 0);
  poke_u32_raw (dhcp->ciaddr, 0);
  poke_u32_raw (dhcp->yiaddr, 0);
  poke_u32_raw (dhcp->siaddr, 0);
  poke_u32_raw (dhcp->giaddr, 0);
  memcpy (dhcp->chaddr, node_eth_addr.bits, ETH_ALEN);
  memset (dhcp->sname+ETH_ALEN, 0, 16-ETH_ALEN);
  memset (dhcp->sname, 0, MAX_SERVER_NAME);
  memset (dhcp->file, 0, MAX_FILE_NAME);
  poke_u32_hton (dhcp->cookie, BOOTP_EXT_COOKIE);
  if (is_discover)
    memcpy (dhcp->options, dhcpdiscover, sizeof dhcpdiscover);
  else
    {
      memcpy (dhcp->options, dhcprequest, sizeof dhcprequest);
      poke_u32_hton (dhcp->options+5, server_id);
      poke_u32_hton (dhcp->options+11, req_addr);
    }
  udp_header_build (p, BOOTP_CLIENT, BOOTP_SERVER,
		    dhcp_size + sizeof (struct udp));
}


#define MAX_STRINGPOOL 1024
static u8 string_pool[MAX_STRINGPOOL];

#define DHCP_XID 0x00004126
int dhcp (struct eth_packet *p, struct dhcp_data *ddp)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct udp *udp = (struct udp *)ip->data;
  struct dhcp *dhcp = (struct dhcp *)udp->data;
  int dhcp_size = sizeof (struct dhcp) - MAX_DHCP_OPTION_LEN;
  int r;
  u32 server_id, req_addr;
  int is_discover = 1;
  u8 *q;
  int n;
  int strindex = 0;
  int len;
  u32 xid, secs;

  xid = DHCP_XID;
  secs = 0;
  server_id = 0;
  req_addr = 0;

  node_ip_addr = 0;
  ddp->my_ip_addr = 0;
  ddp->server_ip_addr = 0;
  ddp->router_ip_addr = 0;
  ddp->dns = 0;
  ddp->netmask = 0;
  ddp->server_name = 0;
  ddp->filename = 0;
  ddp->hostname = 0;
  ddp->rootpath = 0;

 again:
  while (1)
    {
      int i;

      mk_dhcp (p, xid, secs, is_discover, server_id, req_addr);
      transmit_prepare (p);

#define DHCP_TIMEOUT 1000
      i = 0;
      while (1)
	{
	  while ((r = ip_maton (p)) == 1);

	  if (r == 0)
	    {
	      if (peek_u16_ntoh (e->type) != ETHERTYPE_IP)
		continue;

	      if (p->length < ETH_HLEN + sizeof (struct ip) + sizeof (struct udp) + dhcp_size)
		continue;

	      if (peek_u8 (ip->proto) != PROTO_UDP)
		continue;

	      if (peek_u16_raw (udp->checksum) != 0
		  && tcpudp_checksum (ip, PROTO_UDP) != 0)
		continue;

	      if (peek_u16_ntoh (udp->src_port) != BOOTP_SERVER ||
		  peek_u16_ntoh (udp->dst_port) != BOOTP_CLIENT)
		continue;

	      if (peek_u8 (dhcp->op) != BOOTP_REPLY)
		continue;

	      if (peek_u32_ntoh (dhcp->xid) != xid)
		continue;

	      if (is_discover == 0
		  &&
		  !(peek_u8 (dhcp->options) == DHCP_MSG_TYPE
		    && peek_u8 (dhcp->options+1) == 1
		    && peek_u8 (dhcp->options+2) == DHCPACK))
		continue;

	      goto got_packet;
	    }

	  mdelay (1);
	  i++; 
	  if (i >= DHCP_TIMEOUT)
	    break;
	}
    }

 got_packet:
  if (peek_u8 (dhcp->htype) != 1)
    goto again;

  if (peek_u8 (dhcp->hlen) != ETH_ALEN)
    goto again;

  if (peek_u32_ntoh (dhcp->cookie) != BOOTP_EXT_COOKIE)
    goto again;			/* Accept BOOTP (not DHCP) ??? */

  xid = peek_u32_ntoh (dhcp->xid);
  secs = peek_u16_ntoh (dhcp->secs);
  server_id = peek_u32_ntoh (dhcp->siaddr);
  req_addr = peek_u32_ntoh (dhcp->yiaddr);

  if (is_discover == 1)
    {
      if (!(peek_u8 (dhcp->options) == DHCP_MSG_TYPE
	    && peek_u8 (dhcp->options+1) == 1
	    && peek_u8 (dhcp->options+2) == DHCPOFFER))
	xid++;
      else
	{
	  q = dhcp->options+3;
	  while (q < p->data + p->length)
	    {
	      u8 u = peek_u8 (q);

	      if (u == BOOTP_EXT_END)
		break;

	      q++;
	      if (u == BOOTP_EXT_PAD)
		continue;
	      n = peek_u8 (q++);
	      if (u == DHCP_SRV_ID)
		server_id = peek_u32_ntoh (q);
	      q += n;
	    }

	  is_discover = 0;
	}

      goto again;
    }

  /* Got DHCPACK */
  ddp->my_ip_addr = req_addr;
  ddp->server_ip_addr = server_id;

  len = strnlen (dhcp->sname, MAX_SERVER_NAME);
  if (len)
    {
      ddp->server_name = string_pool;
      memcpy (string_pool, dhcp->sname, len);
      string_pool[len] = 0;
      strindex += len + 1;
    }
  len = strnlen (dhcp->file, MAX_FILE_NAME);
  if (len)
    {
      ddp->filename = string_pool + strindex;
      memcpy (string_pool+strindex, dhcp->file, len);
      string_pool[strindex+len] = 0;
      strindex += len +1;
    }

  /* Parse the options */
  q = dhcp->options+3;
  while (q < p->data + p->length)
    {
      u8 u = peek_u8 (q);

      if (u == BOOTP_EXT_END)
	break;

      if (u == BOOTP_EXT_PAD)
	{
	  q++;
	  continue;
	}

      q++;
      n = peek_u8 (q);
      q++;
      if (u == BOOTP_EXT_NETMASK)
	ddp->netmask = peek_u32_ntoh (q);
      else if (u == BOOTP_EXT_ROUTER)
	ddp->router_ip_addr = peek_u32_ntoh (q); /* Only use first */
      else if (u == BOOTP_EXT_DNS)
	ddp->dns = peek_u32_ntoh (q); /* Only use first */
      else if (u == DHCP_SRV_ID)
	server_id = peek_u32_ntoh (q);
      else if (u == BOOTP_EXT_HOSTNAME)
	{
	  if (strindex + n >= MAX_STRINGPOOL)
	    put_string ("dhcp: hostname too long\n"); /* Ignore data */
	  else
	    {
	      ddp->hostname = string_pool + strindex;
	      memcpy (string_pool+strindex, q, n);
	      string_pool[strindex+n] = 0;
	      strindex += n+1;
	    }
	}
      else if (u == BOOTP_EXT_ROOTPATH)
	{
	  if (strindex + n >= MAX_STRINGPOOL)
	    put_string ("dhcp: rootpath too long\n"); /* Ignore data */
	  else
	    {
	      ddp->rootpath = string_pool + strindex;
	      memcpy (string_pool+strindex, q, n);
	      string_pool[strindex+n] = 0;
	      strindex += n+1;
	    }
	}
      else
	/* ignore the option */;

      q += n;
    }

  return 1;
}

#define DNS_SERVER 53
#define DNS_FLAG_QR_RESPONSE	0x8000
#define DNS_FLAG_RD		0x0100
#define DNS_QUERY_TYPE_A	1
#define DNS_QUERY_CLASS_IN	1

struct dns {
  u8 ident[2];
  u8 flags[2];
  u8 num_questions[2];
  u8 num_answers[2];
  u8 num_authority[2];
  u8 num_additional[2];
  u8 data[0];
};

static void mk_dns_query (struct eth_packet *p,
			  const struct eth_addr *dst_eth_addr, u32 dst_ip_addr,
			  const char *hostname,
			  u16 ident, u16 clientport)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct udp *udp = (struct udp *)ip->data;
  struct dns *dns = (struct dns *)udp->data;
  int len = strlen (hostname);
  int dns_size = sizeof (struct dns) + len + 2 + 2 + 2;
  int i;
  int n;

  p->length = ETH_HLEN + sizeof (struct ip) + sizeof (struct udp) + dns_size;
  ether_header_build (p, dst_eth_addr, ETHERTYPE_IP);
  poke_u32_hton (ip->dst, dst_ip_addr);
  ip_header_build (p, PROTO_UDP, IP_TOS_MINIMIZE_DELAY,
		   sizeof (struct ip) + sizeof (struct udp) + dns_size);
  poke_u16_hton (dns->ident, ident);
  poke_u16_hton (dns->flags, DNS_FLAG_RD);
  poke_u16_hton (dns->num_questions, 1);
  poke_u16_hton (dns->num_answers, 0);
  poke_u16_hton (dns->num_authority, 0);
  poke_u16_hton (dns->num_additional, 0);
  memcpy (dns->data+1, hostname, len);
  poke_u8 (dns->data, '.');	/* Sentinel */
  poke_u8 (dns->data+1+len, 0);
  n=0;
  for (i=len; i>=0; i--)
    {
      if (peek_u8 (&dns->data[i]) == '.')
	{
	  poke_u8 (&dns->data[i], n);
	  n = 0;
	}
      else
	n++;
    }
  poke_u16_hton (dns->data+1+len+1, DNS_QUERY_TYPE_A);
  poke_u16_hton (dns->data+1+len+1+2, DNS_QUERY_CLASS_IN);
  udp_header_build (p, clientport, DNS_SERVER, dns_size + sizeof (struct udp));
}

#define DNS_IDENT 0x1234
#define DNS_CLIENT_PORT 16337	/* arbitral */

/* Only support IN A query */
/* Rely on recursive query supported by the server */
int dns_resolv (struct eth_packet *p,
		const struct eth_addr *dst_eth_addr, u32 server_ip_addr,
		const char *hostname, u32 *ipp)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ip *ip = (struct ip *)e->data;
  struct icmp *icmp = (struct icmp *)ip->data;
  struct udp *udp = (struct udp *)ip->data;
  struct dns *dns = (struct dns *)udp->data;

  while (1)
    {
      int i, r;

      mk_dns_query (p, dst_eth_addr, server_ip_addr, hostname,
		    DNS_IDENT, DNS_CLIENT_PORT);
      transmit_prepare (p);

#define DNS_TIMEOUT 1000
      i = 0;
      while (1)
	{
	  while ((r = ip_maton (p)) == 1);

	  if (r == 0)
	    {			/* Got packet */
	      if (peek_u16_ntoh (e->type) != ETHERTYPE_IP)
		continue;

	      if (p->length < ETH_HLEN + sizeof (struct ip) + sizeof (struct udp) + sizeof (struct dns))
		continue;

	      if (peek_u8 (ip->proto) == PROTO_ICMP)
		{			/* Error? */
		  if (peek_u8 (icmp->type) != ICMP_DESTINATION_UNREACHABLE)
		    continue;

		  /* ICMP_REDIRECT? ICMP_PARAMETER_PROBLEM? */
		  /* XXX: Should check if this is for this UDP connection */
		  return -1;
		}

	      if (peek_u8 (ip->proto) != PROTO_UDP ||
		  p->length < ETH_HLEN + peek_u16_ntoh (ip->total_len) ||
		  peek_u16_ntoh (ip->total_len) < sizeof (struct ip) + sizeof (struct udp) + sizeof (struct dns) ||
		  tcpudp_checksum (ip, PROTO_UDP) != 0 ||
		  peek_u32_ntoh (ip->src) != server_ip_addr ||
		  peek_u32_ntoh (ip->dst) != node_ip_addr ||
		  peek_u16_ntoh (udp->src_port) != DNS_SERVER ||
		  peek_u16_ntoh (udp->dst_port) != DNS_CLIENT_PORT)
		continue;


	      if (peek_u16_ntoh (dns->ident) != DNS_IDENT)
		continue;

	      if ((peek_u16_ntoh (dns->flags) & DNS_FLAG_QR_RESPONSE) == 0)
		continue;

	      if (peek_u16_ntoh (dns->num_answers) == 0)
		return -1;	/* Failed */

	      /* OK, we got the answer. */
	      {
		u8 u, *q;

		/* Skip Query data */
		for (q = dns->data; q < p->data + p->length && (u = *q) != 0; )
		  if ((u & 0xc0) == 0xc0) /* Pointer */
		    {
		      q++;
		      break;
		    }
		  else
		    q += u + 1;
		q += 5;

		while (q < p->data + p->length && (u = *q) != 0)
		  if ((u & 0xc0) == 0xc0) /* Pointer */
		    {
		      q++;
		      break;
		    }
		  else
		    q += u + 1;
		q += 9;
		if (peek_u16_ntoh (q) != 4)
		  return -1;
		q += 2;
		*ipp = peek_u32_ntoh (q);

		return 1;
	      }
	    }

	  mdelay (1);
	  i++; 
	  if (i>=DNS_TIMEOUT)
	    break;
	}
    }

  return 1;
}
