/* 
 * Copyright (C) 2005  Free Software Initiative of Japan
 *
 * Author: NIIBE Yutaka <gniibe@fsij.org>
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License version 2 (or later).
 *
 */

#include "types.h"
#include "string.h"
#include "pp.h"
#include "eth.h"
#include "delay.h"
#include "char-io.h"
#include "netinet6.h"
#include "network-order.h"

/* Global Variables here */
struct ipv6_addr node_ipv6_addr;
struct ipv6_addr link_ipv6_addr;

#define IPv6ADDR_PREFIX_BYTES 8

#define MAX_DEST_CACHE 4
static
struct rt_dst_cache
{
  u8 is_valid;
  u8 pad;
  struct eth_addr eth_addr;
  struct ipv6_addr target;
} rt_dst_cache[MAX_DEST_CACHE];

#define MAX_ROUTES 4
/*
 * Two entries are for routes (in case we get redirect or advertisement), 
 * an entry for my network, and an entry for defalut route.
 */
static
struct rt_table
{
  u8 flags;
  u8 prefix;
  struct eth_addr eth_addr;
  struct ipv6_addr target;
  /* XXX: Expire field?*/
} rt_table[MAX_ROUTES];

#define RT_FLAG_VALID  (1<<0)
/* #define RT_FLAG_ (1<<1) */

#define DEFAULT_ROUTE (MAX_ROUTES - 1)

static inline u16 csum_fold (u32 sum)
{
  union { u16 w[2]; u32 l; } x;

  x.l = sum;
  sum = x.w[0] + x.w[1];
  if (sum > 0xffff)
    sum -= 0xffff;

  return ~sum;
}


/* Assume 2-byte alignment */
static u16 in_checksum_sub (void *p, int len, u16 initial)
{
  u16 *wp = (u16 *)p;
  u32 sum = initial;

  while (len >= 2)
    {
      sum += *wp++;
      len -= 2;
    }

  if (len)
    {
      union { u8 c[2]; u16 w; } x;

      x.c[0] =  *(u8 *)wp;
      x.c[1] = 0;
      sum += x.w;
    }

  return csum_fold (sum);
}

static void ether_header_build (struct eth_packet *p,
				const struct eth_addr *dst_eth_addr)
{
  struct eth_frame *e = (struct eth_frame *)p->data;

  memcpy (e->dst, dst_eth_addr->bits, ETH_ALEN);
  memcpy (e->src, node_eth_addr.bits, ETH_ALEN);
  poke_u16_hton (e->type, ETHERTYPE_IPv6);
}


struct ipv6 {
  u8 version_class_flowlabel[4];
  u8 payload_len[2];
  u8 next_header[1];
  u8 hop_limit[1];
  u8 src[16];
  u8 dst[16];
  u8 data[0];
};

#define HOP_BY_HOP       0

#define HOP_LIMIT_DEFAULT  255

#define PROTO_ICMPv6	58
#define PROTO_TCP	 6
#define PROTO_UDP	17

static void ipv6_header_build (struct eth_packet *p, u8 proto, u16 payload_len, u8 hop_limit,
			       const struct ipv6_addr *src, const struct ipv6_addr *dst)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;

  poke_u32_hton (ipv6->version_class_flowlabel, 0x60000000);
  poke_u16_hton (ipv6->payload_len, payload_len);
  poke_u8 (ipv6->next_header, proto);
  poke_u8 (ipv6->hop_limit, hop_limit);
  memcpy (ipv6->src, src->bits, IPv6_ADDRSIZE);
  memcpy (ipv6->dst, dst->bits, IPv6_ADDRSIZE);
}

/* Structure is same as ICMP v4, but checksum calculation is different */
struct icmpv6 {
  u8 type[1];
  u8 code[1];
  u8 checksum[2];
  u8 data[0];
};

#define ICMPv6_ROUTER_SOLICITATION	133
#define ICMPv6_ROUTER_ADVERTISEMENT	134
#define ICMPv6_NC_SOLICITATION		135
#define ICMPv6_NC_ADVERTISEMENT		136
#define ICMPv6_MC_LISTENER_REPORT	143

const u8 multicast_listener_report_message[] = 
  {
    /* IPv6 header */ 
    0x60, 0x00, 0x00, 0x00,
    0, 36, /* Payload length */
    /* Next header is: */ HOP_BY_HOP,
    1, /* Hop limit */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* SRC = :: */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* DST = ff02::16 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x16,

    /* Hop by Hop option */
    /* Next header is: */ PROTO_ICMPv6,
    0, 0x05, 0x02, 0x00, 0x00, 0x01, 0x00, /* Router Aleart: MLD */

    /* ICMPv6 Multicast Listener Report Message v2, to exclude my link multicast */ 
    ICMPv6_MC_LISTENER_REPORT, 0x00, 
    0x00, 0x00, /* Check sum */
    0x00, 0x00, 0x00, 0x01, 0x04, 0x00, 0x00, 0x00, /* Change to exclude */
    0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* My link multicast */ 
    0x00, 0x00, 0x00, 0x01, 0xff, 0x00, 0x00, 0x00, /* ff02::1:ffxx:xxxx */
  };

static inline u32 poke_node_mclink_3bytes (u8 *p)
{
  memcpy (p+13, &node_eth_addr.bits[3], 3);
  return node_eth_addr.bits[3] + (node_eth_addr.bits[4]<<8) + node_eth_addr.bits[5];
}

static void mk_mc_lrm_my_link_multicast (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)&(ipv6->data[8]);
  u32 sum = 0x9075;
  const u8 ether_dst[] = { 0x33, 0x33, 0x00, 0x00, 0x00, 0x16 };

  p->length = ETH_HLEN + sizeof (multicast_listener_report_message);
  ether_header_build (p, (const struct eth_addr *)ether_dst);
  memcpy (e->data, multicast_listener_report_message,
	  sizeof (multicast_listener_report_message));
  sum += poke_node_mclink_3bytes (&icmpv6->data[8]);
  poke_u16_hton (icmpv6->checksum, csum_fold (sum));
}

struct neigh_solicit {
  u8 reserved[4];
  u8 target[16];
  u8 options[0];
};

static void mk_neighbour_solicitation (struct eth_packet *p, const struct ipv6_addr *target)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  struct neigh_solicit *ndisc = (struct neigh_solicit *)icmpv6->data;
  u8 ether_dst[] = { 0x33, 0x33, 0xff, 0x00, 0x00, 0x00 };
  int icmpv6len = 32;
  struct ipv6_addr solicited_node_mc_addr = {
    { 0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
      0x00, 0x00, 0x00, 0x01, 0xff, 0x00, 0x00, 0x00 } };
  u8 options[] = { 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

  p->length = ETH_HLEN + sizeof (struct ipv6) +icmpv6len;
  memcpy (&ether_dst[3], &target->bits[13], 3);
  ether_header_build (p, (struct eth_addr *)ether_dst);

  memcpy (&solicited_node_mc_addr.bits[13], &target->bits[13], 3);
  ipv6_header_build (p, PROTO_ICMPv6, icmpv6len, HOP_LIMIT_DEFAULT, 
		     &link_ipv6_addr, &solicited_node_mc_addr);

  poke_u8 (icmpv6->type, ICMPv6_NC_SOLICITATION);
  poke_u8 (icmpv6->code, 0);
  memset (ndisc->reserved, 0, 4);
  memcpy (ndisc->target, target->bits, IPv6_ADDRSIZE);
  memcpy (&options[2], e->src, ETH_ALEN);
  memcpy (ndisc->options, options, 8);
  poke_u16_raw (icmpv6->checksum, 0);
  poke_u16_raw (icmpv6->checksum,
		in_checksum_sub (ipv6->src,
				 IPv6_ADDRSIZE*2 + icmpv6len,
				 htons (icmpv6len) + htons (PROTO_ICMPv6)));
}

u8 neighbor_solicitation_my_link[] = 
  {
    /* IPv6 header */ 
    0x60, 0x00, 0x00, 0x00,
    0, 24, /* Payload length */
    PROTO_ICMPv6,
    HOP_LIMIT_DEFAULT,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* SRC = :: */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* DST = ff02::1:ffxx:xxxx */
    0x00, 0x00, 0x00, 0x01, 0xff, 0x00, 0x00, 0x00,

    /* ICMPv6 neighbor solicitation */ 
    ICMPv6_NC_SOLICITATION, 0x00, 
    0x00, 0x00, /* Check sum */
    0x00, 0x00, 0x00, 0x00,
    0xfe, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* My link address */ 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* fe80::xxxx:xxff:fexx:xxxx */
  };

static inline u32 poke_link_eui (u8 *p)
{
  poke_u16_hton (p+8, 0x0200 | (node_eth_addr.bits[0]<<8) | node_eth_addr.bits[1]);
  poke_u8 (p+10, node_eth_addr.bits[2]);
  poke_u8 (p+11, 0xff);
  poke_u8 (p+12, 0xfe);
  memcpy (p+13, &node_eth_addr.bits[3], 3);

  return (0x0200 | (node_eth_addr.bits[0]<<8)) + node_eth_addr.bits[1]
    + (node_eth_addr.bits[2]<<8) + node_eth_addr.bits[3]
    + (node_eth_addr.bits[4]<<8) + node_eth_addr.bits[5] + 0xfffe;
}

static void mk_ns_my_link (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  u32 sum = 0x82d9;
  u8 ether_dst[] = { 0x33, 0x33, 0xff, 0x00, 0x00, 0x00 };

  p->length = ETH_HLEN + sizeof (neighbor_solicitation_my_link);
  memcpy (&ether_dst[3], &node_eth_addr.bits[3], 3);
  ether_header_build (p, (struct eth_addr *)ether_dst);
  memcpy (e->data, neighbor_solicitation_my_link, sizeof (neighbor_solicitation_my_link));
  sum += poke_node_mclink_3bytes (ipv6->dst);
  sum += poke_link_eui (&icmpv6->data[4]);
  poke_u16_hton (icmpv6->checksum, csum_fold (sum));
}

static inline u32 poke_link_addr (u8 *p)
{
  memcpy (p, node_eth_addr.bits, ETH_ALEN);

  return (node_eth_addr.bits[0]<<8) + node_eth_addr.bits[1]
    + (node_eth_addr.bits[2]<<8) + node_eth_addr.bits[3]
    + (node_eth_addr.bits[4]<<8) + node_eth_addr.bits[5];
}

u8 router_solicitation[] = 
  {
    /* IPv6 header */ 
    0x60, 0x00, 0x00, 0x00,
    0, 16, /* Payload length */
    PROTO_ICMPv6,
    HOP_LIMIT_DEFAULT,
    0xfe, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* SRC = My link address */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* fe80::xxxx:xxff:fexx:xxxx */
    0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* DST = All routers */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, /* ff02::2 */

    /* ICMPv6 router solicitation */ 
    ICMPv6_ROUTER_SOLICITATION, 0x00, 
    0x00, 0x00, /* Check sum */
    0x00, 0x00, 0x00, 0x00, 0x01, 0x01,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* Ether address */ 
  };

static void mk_rs (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  u32 sum = 0x82d2;
  const u8 ether_dst[] = { 0x33, 0x33, 0xff, 0x00, 0x00, 0x02 }; /* All Routers */

  p->length = ETH_HLEN + sizeof (router_solicitation);
  ether_header_build (p, (const struct eth_addr *)ether_dst);
  memcpy (e->data, router_solicitation, sizeof (router_solicitation));

  sum += poke_link_eui (ipv6->src);
  sum += poke_link_addr (&icmpv6->data[6]);
  poke_u16_hton (icmpv6->checksum, csum_fold (sum));
}

static inline u32 poke_node_ipv6_addr (u8 *p)
{
  memcpy (p, node_ipv6_addr.bits, IPv6_ADDRSIZE);
  return ((node_ipv6_addr.bits[0] + node_ipv6_addr.bits[2] + node_ipv6_addr.bits[4]
    + node_ipv6_addr.bits[6] + node_ipv6_addr.bits[8] + node_ipv6_addr.bits[10]
    + node_ipv6_addr.bits[12] + node_ipv6_addr.bits[14]) << 8)
    + node_ipv6_addr.bits[1] + node_ipv6_addr.bits[3] + node_ipv6_addr.bits[5]
    + node_ipv6_addr.bits[7] + node_ipv6_addr.bits[9] + node_ipv6_addr.bits[11]
    + node_ipv6_addr.bits[13] + node_ipv6_addr.bits[15];
}

u8 neighbor_solicitation_my_global[] = 
  {
    /* IPv6 header */ 
    0x60, 0x00, 0x00, 0x00,
    0, 24, /* Payload length */
    PROTO_ICMPv6,
    HOP_LIMIT_DEFAULT,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* SRC = :: */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* DST = ff02::1:ffxx:xxxx */
    0x00, 0x00, 0x00, 0x01, 0xff, 0x00, 0x00, 0x00,

    /* ICMPv6 neighbor solicitation */ 
    ICMPv6_NC_SOLICITATION, 0x00, 
    0x00, 0x00, /* Check sum */
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* My global address */ 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };

static void mk_ns_my_global (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  u32 sum = 0x8557;
  u8 ether_dst[] = { 0x33, 0x33, 0xff, 0x00, 0x00, 0x00 };

  p->length = ETH_HLEN + sizeof (neighbor_solicitation_my_global);
  memcpy (&ether_dst[3], &node_eth_addr.bits[3], 3);
  ether_header_build (p, (const struct eth_addr *)ether_dst);
  memcpy (e->data, neighbor_solicitation_my_global,
	  sizeof (neighbor_solicitation_my_global));
  sum += poke_node_mclink_3bytes (ipv6->dst);
  sum += poke_node_ipv6_addr (&icmpv6->data[4]);
  poke_u16_hton (icmpv6->checksum, csum_fold (sum));
}

struct icmpv6_router_advertisement {
  u8 type[1];
  u8 code[1];
  u8 checksum[2];
  u8 max_hop_limit[1];
  u8 flags[1];
  u8 router_lifetime[2];
  u8 rt[4];
  u8 rrt[4];
  u8 data[0];
};

#define ICMPv6_OPTION_SOURCE_LINK_LAYER 1
#define ICMPv6_OPTION_PREFIX_INFO      3

struct icmpv6_option_ll_addr {
  u8 type[1];
  u8 len[1];
  u8 lladdr[6];
};

struct icmpv6_option_prefix_info {
  u8 type[1];
  u8 len[1];
  u8 prefix_len[1];
  u8 flags[1];
  u8 valid_lifetime[4];
  u8 preferred_lifetime[4];
  u8 reserved[4];
  u8 prefix[16];
};

/**
 * net6_address_autoconfig - Configure interface using automatic
 *                           stateless IPv6 configuration mechanism
 * 
 */
int
net6_address_autoconfig (struct eth_packet *p)
{
  int r;
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  int prefix_len;

  /* Assume link_ipv6_addr is zero cleared. */
  link_ipv6_addr.bits[0] = 0xfe;
  link_ipv6_addr.bits[1] = 0x80;
  poke_link_eui (link_ipv6_addr.bits);

  mk_mc_lrm_my_link_multicast (p);
  transmit_prepare (p);
  while ((r = eth_maton (p)) == 1); /* Busy, wait until done. */
  /* XXX: duplication check should come here */

  put_string ("Multicast listener report message ... done.\n");

  mk_ns_my_link (p);
  transmit_prepare (p);
  while ((r = eth_maton (p)) == 1); /* Busy, wait until done. */
  /* XXX: duplication check should come here */

  put_string ("Neighbour solicitation for my own link address ... done.\n");

  while (1)
    {
      int i;

      mk_rs (p);
      transmit_prepare (p);

      put_string ("Send a 'Router Solicitation'.\n");

#define RS_TIMEOUT 1000
      for (i = 0; i < RS_TIMEOUT; i++)
	{
	  while ((r = eth_maton (p)) == 1); /* Busy, wait until done. */

	  if (r == 0)
	    {
	      put_string ("Got a 'Router Advertisement'!\n");

	      if (peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
		continue;

	      if (peek_u8 (ipv6->next_header) != PROTO_ICMPv6)
		continue;

	      if (p->length < (ETH_HLEN + sizeof (struct ipv6) + sizeof (struct icmpv6)
			       + 6 + 16 + 6 + 16 + 2 + 6)
		  || p->length < (ETH_HLEN + sizeof (struct ipv6)
				  + peek_u16_ntoh (ipv6->payload_len)))
		continue;

	      if (in_checksum_sub (ipv6->src,
				   IPv6_ADDRSIZE*2
				   + peek_u16_ntoh (ipv6->payload_len), 
				   peek_u16_raw (ipv6->payload_len)
				   + htons (PROTO_ICMPv6)) != 0)
		continue;

	      /* Got it! */
	      {
		struct icmpv6_router_advertisement *ra = (struct icmpv6_router_advertisement *)icmpv6;
		int len = peek_u16_ntoh (ipv6->payload_len) - sizeof (struct icmpv6_router_advertisement);
		int i;
		struct icmpv6_option_prefix_info *prefix_info = 0;

		for (i = 0; i < len;)
		  if (ra->data[i] == ICMPv6_OPTION_PREFIX_INFO)
		    {
		      prefix_info = (struct icmpv6_option_prefix_info *)&ra->data[i];
		      break;
		    }
		  else
		    i += ra->data[i+1]*8;

		if (!prefix_info)
		  {
		    put_string ("  *** Not good.\n");
		    continue;
		  }

		/* Assume prefixlen is multiple of 8 */
		prefix_len = peek_u8(prefix_info->prefix_len);
		if (prefix_len != IPv6ADDR_PREFIX_BYTES * 8)
		  {
		    put_string ("  *** Not good prefix (!= 64).\n");
		    continue;
		  }

		memcpy (node_ipv6_addr.bits, prefix_info->prefix, IPv6ADDR_PREFIX_BYTES);
		goto got_address_prefix;
	      }
	    }

	  mdelay (1);
	}
    }
 got_address_prefix:
  /* Put the router's ethernet address into routing table. */
  rt_table[DEFAULT_ROUTE].flags = RT_FLAG_VALID;
  rt_table[DEFAULT_ROUTE].prefix = 0;
  memcpy (rt_table[DEFAULT_ROUTE].eth_addr.bits, e->src, ETH_ALEN);

  /* My IP address */
  /* XXX more sanity check please... Is autoconf OK?*/
  poke_link_eui (node_ipv6_addr.bits);

  mk_ns_my_global (p);
  transmit_prepare (p);
  while ((r = eth_maton (p)) == 1); /* Busy, wait until done. */
  /* XXX: here comes duplicate detection */

  return 0;
}


#define ICMPv6_ECHO_REQUEST		128
#define ICMPv6_ECHO_REPLY		129

#define ICMP_DESTINATION_UNREACHABLE	3
#define ICMP_SOURCE_QUENCH		4
#define ICMP_REDIRECT			5
#define ICMP_PARAMETER_PROBLEM		12

struct icmp_echo {
  u8 id[2];
  u8 seq[2];
  u8 data[0];
};

static void mk_icmpv6_echo (struct eth_packet *p,
			    const struct eth_addr *dst_eth_addr,
			    const struct ipv6_addr *dst_ipv6_addr_p,
			    unsigned int icmp_echo_id, unsigned int seq,
			    unsigned char *data, int len)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  struct icmp_echo *icmp_echo = (struct icmp_echo *)icmpv6->data;
  int icmpv6len = sizeof (struct icmpv6) + sizeof (struct icmp_echo) + len;


  p->length = ETH_HLEN + sizeof (struct ipv6) +icmpv6len ;
  ether_header_build (p, dst_eth_addr);

  ipv6_header_build (p, PROTO_ICMPv6, icmpv6len, HOP_LIMIT_DEFAULT,
		     &node_ipv6_addr, dst_ipv6_addr_p);

  poke_u8 (icmpv6->type, ICMPv6_ECHO_REQUEST);
  poke_u8 (icmpv6->code, 0);
  poke_u16_raw (icmpv6->checksum, 0);
  poke_u16_hton (icmp_echo->id, icmp_echo_id);
  poke_u16_hton (icmp_echo->seq, seq);
  memcpy (icmp_echo->data, data, len);
  poke_u16_raw (icmpv6->checksum,
		in_checksum_sub (ipv6->src,
				 IPv6_ADDRSIZE*2 + icmpv6len,
				 htons (icmpv6len) + htons (PROTO_ICMPv6)));
}

static int handle_icmpv6_echo_reply (struct eth_packet *p,
				     const struct ipv6_addr *ipv6_addr_p, int seq)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  struct icmp_echo *icmp_echo = (struct icmp_echo *)icmpv6->data;

  if (peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
    return 0;

  if (peek_u8 (ipv6->next_header) != PROTO_ICMPv6)
    return 0;

  if (peek_u8 (icmpv6->type) != ICMPv6_ECHO_REPLY)
    return 0;

  if (p->length < ETH_HLEN + sizeof (struct ipv6) + sizeof (struct icmpv6)
      + sizeof (struct icmp_echo))
    return 0;

  if (peek_u16_ntoh (icmp_echo->seq) != seq)
    return 0;

  return 1;
}

static int icmpv6_echo_request_serv (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  int icmpv6len;

  if (p->length < ETH_HLEN + sizeof (struct ipv6) + sizeof (struct icmpv6)
      + sizeof (struct icmp_echo))
    return -1;			/* Discard */

  icmpv6len = peek_u16_ntoh (ipv6->payload_len);
  if (p->length < ETH_HLEN + sizeof (struct ipv6) + icmpv6len
      || in_checksum_sub (ipv6->src, IPv6_ADDRSIZE*2 + icmpv6len,
			  htons (icmpv6len) + htons (PROTO_ICMPv6)) != 0)
    return -1;			/* Discard */

  /* Build a reply packet */
  p->length = ETH_HLEN + sizeof (struct ipv6) + icmpv6len;
  ether_header_build (p, (struct eth_addr *)e->src);

  memcpy (ipv6->dst, ipv6->src, IPv6_ADDRSIZE);
  memcpy (ipv6->src, node_ipv6_addr.bits, IPv6_ADDRSIZE);
  poke_u8 (icmpv6->type, ICMPv6_ECHO_REPLY);
  poke_u16_raw (icmpv6->checksum, 0);
  poke_u16_raw (icmpv6->checksum,
		in_checksum_sub (ipv6->src, IPv6_ADDRSIZE*2 + icmpv6len,
				 htons (icmpv6len) + htons (PROTO_ICMPv6)));
  transmit_prepare (p);
  return 1;
}

struct neigh_adv {
  u8 RSO_and_reserved[4];
  u8 target[16];
  u8 options[0];
};

static int ndisc_solicitation_serv (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  struct neigh_solicit *neigh_solicit = (struct neigh_solicit *)icmpv6->data;
  struct neigh_adv *neigh_adv = (struct neigh_adv *)icmpv6->data;
  int icmpv6len;
  u8 dst[] = { 0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	       0x00, 0x00, 0x00, 0x01, 0xff, 0x00, 0x00, 0x00};
  u8 options[] = { 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

  memcpy (&dst[13], &node_ipv6_addr.bits[13], 3); /* NDISC solicitation multicast address */

  if (memcmp (dst, ipv6->dst, IPv6_ADDRSIZE) != 0
      && memcmp (link_ipv6_addr.bits, ipv6->dst, IPv6_ADDRSIZE) != 0)
    return -1;			/* Discard */

  if (p->length < (ETH_HLEN + sizeof (struct ipv6)
		   + sizeof (struct icmpv6) + sizeof (struct neigh_solicit)))
    /* Discard */
    return -1;

  /* XXX: More sanity check... */

  if (memcmp (node_ipv6_addr.bits, neigh_solicit->target, IPv6_ADDRSIZE) != 0
      && memcmp (link_ipv6_addr.bits, neigh_solicit->target, IPv6_ADDRSIZE) != 0)
    /* Discard */
    return -1;

  /* Build a reply (advertisement) packet */
  icmpv6len = 32;
  p->length = ETH_HLEN + sizeof (struct ipv6) + icmpv6len;
  ether_header_build (p, (struct eth_addr *)e->src);

  memcpy (ipv6->dst, ipv6->src, IPv6_ADDRSIZE);
  memcpy (ipv6->src, link_ipv6_addr.bits, IPv6_ADDRSIZE);
  poke_u8 (icmpv6->type, ICMPv6_NC_ADVERTISEMENT);
  poke_u8 (icmpv6->code, 0);
  poke_u8 (&neigh_adv->RSO_and_reserved[0], 0x60); /* Not router, Solicited, Override */
  poke_u8 (&neigh_adv->RSO_and_reserved[1], 0x0);
  poke_u8 (&neigh_adv->RSO_and_reserved[2], 0x0);
  poke_u8 (&neigh_adv->RSO_and_reserved[3], 0x0);
  /* neigh_adv->target is as is. */
  memcpy (&options[2], e->src, ETH_ALEN);
  memcpy (neigh_adv->options, options, 8);
  poke_u16_raw (icmpv6->checksum, 0);
  poke_u16_raw (icmpv6->checksum,
		in_checksum_sub (ipv6->src, IPv6_ADDRSIZE*2 + icmpv6len,
				 htons (icmpv6len) + htons (PROTO_ICMPv6)));
  transmit_prepare (p);
  return 1;
}

static int ipv6_maton (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  int icmpv6len;
  int r;

  if ((r = eth_maton (p)) != 0)
    return r;

  if (peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
    return -1;

  if (p->length < ETH_HLEN + sizeof (struct ipv6)
      /* XXX V6 header version check should come here */)
    return -1;

  if (peek_u8 (ipv6->next_header) != PROTO_ICMPv6)
    return 0;

  icmpv6len = peek_u16_ntoh (ipv6->payload_len);

  if (p->length < ETH_HLEN + sizeof (struct ipv6) + icmpv6len
      || in_checksum_sub (ipv6->src, IPv6_ADDRSIZE*2 + icmpv6len,
			  htons (icmpv6len) + htons (PROTO_ICMPv6)) != 0)
    /* Discard */
    return -1;

  if (peek_u8 (icmpv6->type) == ICMPv6_NC_SOLICITATION)
    {				/* Handle neighbour solicitation */
      /* XXX: sanity check needed: is it through ns-multicast address? */
      return ndisc_solicitation_serv (p);
    }

  /* XXX: Routing here */
  /* XXX: Router advertisement */
  /* XXX: Redirect */
  if (peek_u8 (icmpv6->type) == ICMPv6_ROUTER_ADVERTISEMENT)
    {
    }

  /* From here, unicast is only valid */
  if (memcmp (node_ipv6_addr.bits, ipv6->dst, IPv6_ADDRSIZE) != 0
      && memcmp (link_ipv6_addr.bits, ipv6->dst, IPv6_ADDRSIZE) != 0)
    /* Discard */
    return -1;

  if (peek_u8 (icmpv6->type) != ICMPv6_ECHO_REQUEST)
    return 0;

  r = icmpv6_echo_request_serv (p);
  return r;
}

static int
handle_ndisc_reply (struct eth_packet *p, const struct ipv6_addr *target,
		    struct eth_addr *eth_addr)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  int icmpv6len;

  if (peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
    return 0;

  if (peek_u8 (ipv6->next_header) != PROTO_ICMPv6)
    return 0;

  if (peek_u8 (icmpv6->type) != ICMPv6_NC_ADVERTISEMENT)
    return 0;

  icmpv6len = peek_u16_ntoh (ipv6->payload_len);

  if (p->length < ETH_HLEN + sizeof (struct ipv6) + sizeof (struct icmpv6)
      + sizeof (struct neigh_adv)
      || p->length < ETH_HLEN + sizeof (struct ipv6) + icmpv6len)
    return 0;

  if (memcmp (ipv6->dst, link_ipv6_addr.bits, IPv6_ADDRSIZE) != 0
      || memcmp (ipv6->src, target->bits, IPv6_ADDRSIZE) != 0)
    return 0;

  /* XXX: More sanity check??? */
  memcpy (eth_addr->bits, e->src, ETH_ALEN);
  return 1;
}

/*
 * Three of destination cache
 * One for this ethernet
 * One for default router
 */
#define MAX_NDISC_TRY 3
int ndisc_resolv (struct eth_packet *p, const struct ipv6_addr *target,
		  struct eth_addr *dst_eth_addr)
{
  int try;

  for (try=0; try < MAX_NDISC_TRY; try++)
    {
      int i, r;

      mk_neighbour_solicitation (p, target);
      transmit_prepare (p);

#define NDISC_TIMEOUT 500
      for (i = 0; i < NDISC_TIMEOUT; i++) 
	{
	  while ((r = ipv6_maton (p)) == 1);

	  if (r == 0 && handle_ndisc_reply (p, target, dst_eth_addr) != 0)
	    return 1;

	  mdelay (1);
	}
    }

  return -1;
}

struct eth_addr *ipv6_get_route (struct eth_packet *p, const struct ipv6_addr *target)
{
  int i;
  int hash = target->bits[IPv6_ADDRSIZE-1] % MAX_DEST_CACHE;

  /* Check destination cache */
  if (rt_dst_cache[hash].is_valid != 0
      && memcmp (rt_dst_cache[hash].target.bits, target->bits, IPv6_ADDRSIZE) == 0)
    return &rt_dst_cache[hash].eth_addr;

  /* Check this network */
  if (memcmp (node_ipv6_addr.bits, target->bits, IPv6ADDR_PREFIX_BYTES) == 0)
    { /* Target is on this network, let's proceed to neighbour discovery. */
      struct eth_addr target_eth_addr;
      if (ndisc_resolv (p, target, &target_eth_addr) < 0)
	return 0;		/* Failure */
	    
      rt_dst_cache[hash].is_valid = 1;
      memcpy (rt_dst_cache[hash].target.bits, target->bits, IPv6_ADDRSIZE);
      memcpy (rt_dst_cache[hash].eth_addr.bits, target_eth_addr.bits, ETH_ALEN);
      return &rt_dst_cache[hash].eth_addr;
    }

  for (i = 0; i < MAX_ROUTES; i++)
    {
      if ((rt_table[i].flags & RT_FLAG_VALID) == 0)
	continue;

      /* Here, we assume prefix is multiple of 8-bit */
      if (memcmp (rt_table[i].target.bits, target->bits, rt_table[i].prefix >> 3) == 0)
	{
	  rt_dst_cache[hash].is_valid = 1;
	  memcpy (rt_dst_cache[hash].target.bits, target->bits, IPv6_ADDRSIZE);
	  memcpy (rt_dst_cache[hash].eth_addr.bits, rt_table[i].eth_addr.bits, ETH_ALEN);
	  return &rt_dst_cache[hash].eth_addr;
	}
    }

  return 0;			/* Failed */
}

int ping6 (struct eth_packet *p, const struct eth_addr *dst_eth_addr, 
	   const struct ipv6_addr *dst_ipv6_addr_p, int count)
{
  int seq = 0;

  while (seq < count)
    {
      int i, r;

      mk_icmpv6_echo (p, dst_eth_addr, dst_ipv6_addr_p, 0xbeef, seq,
		      "Hello IPv6! Hello IPv6!! Hello IPv6!!!", 38);
      transmit_prepare (p);

#define PING_TIMEOUT 1000
      for (i = 0; i < PING_TIMEOUT; i++)
	{
	  while ((r = ipv6_maton (p)) == 1);

	  if (r == 0 && handle_icmpv6_echo_reply (p, dst_ipv6_addr_p, seq) != 0)
	    return 1;

	  mdelay (1);
	}

      seq++;
    }

  return -1;
}

static u16 tcpudp_checksum (struct ipv6 *ipv6, u8 proto)
{
  int len, orig_len;

  len = orig_len = peek_u16_ntoh (ipv6->payload_len);
  if ((len & 1))
    {
      poke_u8 (&ipv6->data[len], 0);
      len++;
    }

  return in_checksum_sub (ipv6->src, IPv6_ADDRSIZE*2+len, htons (orig_len) + htons (proto));
}

#define TCP_URG		0x20
#define TCP_ACK		0x10
#define TCP_PSH		0x08
#define TCP_RST		0x04
#define TCP_SYN		0x02
#define TCP_FIN		0x01

#define TCP_ISN		0x3e33268f

struct tcp {
  u8 src_port[2];
  u8 dst_port[2];
  u8 seq[4];
  u8 ack[4];
  u8 hlen[1];
  u8 flags[1];
  u8 window_size[2];
  u8 checksum[2];
  u8 urgent[2];
  u8 data[0];
};

static void tcp_header_build (struct eth_packet *p,
			      u32 seq, u32 ack, u8 hlen, u8 flags, u16 w_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct tcp *tcp = (struct tcp *)ipv6->data;

  poke_u32_hton (tcp->seq, seq);
  poke_u32_hton (tcp->ack, ack);
  poke_u8 (tcp->hlen, hlen<<2);
  poke_u8 (tcp->flags, flags);
  poke_u16_hton (tcp->window_size, w_size);
  poke_u16_raw (tcp->checksum, 0);
  poke_u16_hton (tcp->urgent, 0);
  poke_u16_raw (tcp->checksum, tcpudp_checksum (ipv6, PROTO_TCP));
}

#define TCP_OPTION_END  0
#define TCP_OPTION_NOP  1
#define TCP_OPTION_MSS	2

static void
mk_tcp_syn (struct eth_packet *p,
	    const struct eth_addr *dst_eth_addr,
	    const struct ipv6_addr *dst_ipv6_addr_p, u16 dst_port,
	    u16 my_port, u16 mss, u32 seq, u16 window_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct tcp *tcp = (struct tcp *)ipv6->data;

  p->length = ETH_HLEN + sizeof (struct ipv6) + sizeof (struct tcp)
    + 4; /* Four bytes for TCP mss option */
  ether_header_build (p, dst_eth_addr);
  ipv6_header_build (p, PROTO_TCP, sizeof (struct tcp) + 4, HOP_LIMIT_DEFAULT,
		     &node_ipv6_addr, dst_ipv6_addr_p);
  poke_u16_hton (tcp->src_port, my_port);
  poke_u16_hton (tcp->dst_port, dst_port);
  tcp->data[0] = TCP_OPTION_MSS;
  tcp->data[1] = 4;
  poke_u16_hton (&tcp->data[2], mss);
  tcp_header_build (p, seq, 0, 24, TCP_SYN, window_size);
}

static int tcp_syn_ack_serv (struct eth_packet *p, unsigned int t_seq,
			     u8 *send, int send_size, int window_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct tcp *tcp = (struct tcp *)ipv6->data;
  u16 port;
  u32 seq;
  struct ipv6_addr dst;

  seq = peek_u32_ntoh (tcp->seq);

  if (peek_u16_ntoh (tcp->window_size) < send_size)
    put_string ("eth: sending TCP data too big!\n"); /* XXX should do something */

  p->length = ETH_HLEN + send_size + sizeof (struct ipv6) + sizeof (struct tcp);
  ether_header_build (p, (struct eth_addr *)e->src);
  memcpy (dst.bits, ipv6->src, IPv6_ADDRSIZE);
  ipv6_header_build (p, PROTO_TCP, sizeof (struct tcp) + send_size, HOP_LIMIT_DEFAULT,
		     &node_ipv6_addr, &dst);
  port = peek_u16_raw (tcp->src_port);
  poke_u16_raw (tcp->src_port, peek_u16_raw (tcp->dst_port));
  poke_u16_raw (tcp->dst_port, port);
  memcpy (tcp->data, send, send_size);
  tcp_header_build (p, t_seq, seq+1, 20, TCP_ACK|TCP_PSH|TCP_FIN, window_size);

  transmit_prepare (p);
  return seq + 1; /* SYN consumes 1-byte. */
}

static void tcp_reset (struct eth_packet *p,
		       const struct eth_addr *dst_eth_addr,
		       const struct ipv6_addr *dst_ipv6_addr_p,
		       u16 dst_port, u16 my_port, u32 seq)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct tcp *tcp = (struct tcp *)ipv6->data;

  p->length = ETH_HLEN + sizeof (struct ipv6) + sizeof (struct tcp);
  ether_header_build (p, dst_eth_addr);
  ipv6_header_build (p, PROTO_TCP, sizeof (struct tcp), HOP_LIMIT_DEFAULT, 
		     &node_ipv6_addr, dst_ipv6_addr_p);
  poke_u16_hton (tcp->src_port, my_port);
  poke_u16_hton (tcp->dst_port, dst_port);
  tcp_header_build (p, seq, 0, 20, TCP_RST, 0xffff);

  transmit_prepare (p);
}

static int tcp_ack (struct eth_packet *p, u32 t_seq, u32 r_seq, u16 window_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct tcp *tcp = (struct tcp *)ipv6->data;
  u16 port;
  int r = 0;
  struct ipv6_addr dst;

  p->length = ETH_HLEN + sizeof (struct ipv6) + sizeof (struct tcp);
  ether_header_build (p, (struct eth_addr *)e->src);
  memcpy (dst.bits, ipv6->src, IPv6_ADDRSIZE);
  ipv6_header_build (p, PROTO_TCP, sizeof (struct tcp), HOP_LIMIT_DEFAULT, 
		     &node_ipv6_addr, &dst);
  port = peek_u16_raw (tcp->src_port);
  poke_u16_raw (tcp->src_port, peek_u16_raw (tcp->dst_port));
  poke_u16_raw (tcp->dst_port, port);
  tcp_header_build (p, t_seq, r_seq, 20, TCP_ACK, window_size);

  transmit_prepare (p);
  return r;
}

#define CONNECTION_TIMEOUT	1000
#define CONNECTION_RETRY	5
#define RETRANSMIT_TIMEOUT	100
#define TIMEWAIT_TIMEOUT	1000

#define IP_LOOP_COUNT  50	/* 50 times nealy equals to 1ms  */

#define ANNOUNCED_MSS		(ETHER_FRAME_LEN-ETH_HLEN-60) /* IPv6=40 TCP=20 bytes long */

#define WINDOW_SIZE	6000

int tcp_maton6 (struct eth_packet *p,
		const struct eth_addr *dst_eth_addr,
		const struct ipv6_addr *server_ipv6_addr_p,
		u16 server_port, u16 my_port,
		u8 *send, int send_size, u8 *receive, int receive_size)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  struct tcp *tcp = (struct tcp *)ipv6->data;
  int r;
  unsigned int count, retransmit_time;
  int retry = 0;
  u32 t_seq = TCP_ISN, r_seq, seq;
  u16 mss = 536;
  u8 c;
  u16 w;
  u32 x;
  int i;
  int len, offset;
  int receive_count = 0;
  int window_size = WINDOW_SIZE; /* Because of ethernet layer (which has 4*1.5KB buffer) */

  /* Connect */
  mk_tcp_syn (p, dst_eth_addr, server_ipv6_addr_p, server_port,
	      my_port, ANNOUNCED_MSS, t_seq, window_size);
  transmit_prepare (p);

  count = 1;
  retransmit_time = CONNECTION_TIMEOUT;
  while (1)
    {
      if (count && (int)(retransmit_time - count) < 0) /* Timer expired */
	{
	  mk_tcp_syn (p, dst_eth_addr, server_ipv6_addr_p, server_port,
		      my_port, ANNOUNCED_MSS, t_seq, window_size);
	  transmit_prepare (p);

	  ++retry;
	  retransmit_time = (CONNECTION_TIMEOUT<<retry) + count;
	  if (retry > CONNECTION_RETRY)
	    /* Really timeout */
	    return -1;
	}

      i = 0;
      while ((r = ipv6_maton (p)) == 1)
	i++;
      count += i / IP_LOOP_COUNT;

      if (r < 0 || peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
	goto next;

      if (peek_u8 (ipv6->next_header) == PROTO_ICMPv6)
	{			/* Error? */
	  if (peek_u8 (icmpv6->type) != ICMP_DESTINATION_UNREACHABLE)
	    goto next;

	  /* ICMP_REDIRECT? ICMP_PARAMETER_PROBLEM? */
	  /* XXX: Should check if this is for this TCP connection */
	  return -1;
	}

      if (peek_u8 (ipv6->next_header) != PROTO_TCP
	  || p->length < ETH_HLEN + sizeof (struct ipv6) + peek_u16_ntoh (ipv6->payload_len)
	  || peek_u16_ntoh (ipv6->payload_len) < sizeof (struct tcp)
	  || tcpudp_checksum (ipv6, PROTO_TCP) != 0
	  || memcmp (ipv6->src, server_ipv6_addr_p->bits, IPv6_ADDRSIZE) != 0
	  || memcmp (ipv6->dst, node_ipv6_addr.bits, IPv6_ADDRSIZE) != 0
	  || peek_u16_ntoh (tcp->src_port) != server_port
	  || peek_u16_ntoh (tcp->dst_port) != my_port)
	goto next;

      if ((peek_u8 (tcp->flags) & TCP_RST))
	return -1;

      break;

    next:
      mdelay (1);
      if (count)
	count++;
    }

  t_seq++;
 syn_ack:
  if ((peek_u8 (tcp->flags) & (TCP_SYN|TCP_ACK)) != (TCP_SYN|TCP_ACK))
    {
      tcp_reset (p, dst_eth_addr, server_ipv6_addr_p, server_port, my_port, t_seq);
      return -1; /* XXX: upper layer have to call eth_maton to send RST...  */
    }

  if (peek_u32_ntoh (tcp->ack) != t_seq)
    {
      tcp_reset (p, dst_eth_addr, server_ipv6_addr_p, server_port, my_port, t_seq);
      return -1; /* XXX: upper layer have to call eth_maton to send RST...  */
    }

  if ((c = peek_u8 (tcp->hlen) >> 2) > sizeof (struct tcp))
    {				/* TCP option */
      int i;
      int len = c - sizeof (struct tcp);
      int data_len;

      /* We only handle MSS */
      for (i=0; i<len; i+= data_len)
	if ((c = peek_u8 (&tcp->data[i])) == TCP_OPTION_MSS
	    && peek_u8 (&tcp->data[i+1]) == 4)
	  {
	    mss = peek_u16_ntoh (&tcp->data[2]);
	    break;
	  }
	else if (c == TCP_OPTION_END)
	  break;
	else if (c == TCP_OPTION_NOP)
	  data_len = 1;
	else
	  data_len = peek_u8 (&tcp->data[i+1]);
    }

  if (mss < send_size)
    put_string ("eth: sending TCP data too big!\n"); /* XXX should do something */

  /* Connection established */
  r_seq = tcp_syn_ack_serv (p, t_seq, send, send_size, window_size);
  t_seq += send_size + 1;	/* FIN consumes 1-byte. */

  /* FIN-WAIT-1 */
  count = 0;
  while (1)
    {
      if (count && (int)(retransmit_time - count) < 0) /* Timer expired */
	break;

      i = 0;
      while ((r = ipv6_maton (p)) == 1)
	i++;
      count += i / IP_LOOP_COUNT;

      if (r < 0)
	{ /* When we go to idle state, reset window size to max. */
	  if (p->state == ETH_IDLE && !p->tx_completion_wait)
	    window_size = WINDOW_SIZE;
	  goto next_1;
	}

      if (peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
	goto next_1;		/* Discard */

      if (peek_u8 (ipv6->next_header) == PROTO_ICMPv6)
	{			/* Error? */
	  if (peek_u8 (icmpv6->type) != ICMP_DESTINATION_UNREACHABLE)
	    goto next_1;

	  /* ICMP_REDIRECT? ICMP_PARAMETER_PROBLEM? */
	  /* XXX: Should check if this is for this TCP connection */
	  return -1;
	}

      if (peek_u8 (ipv6->next_header) != PROTO_TCP
	  || p->length < ETH_HLEN + sizeof (struct ipv6) + peek_u16_ntoh (ipv6->payload_len)
	  || peek_u16_ntoh (ipv6->payload_len) < sizeof (struct tcp)
	  || tcpudp_checksum (ipv6, PROTO_TCP) != 0
	  || memcmp (ipv6->src, server_ipv6_addr_p->bits, IPv6_ADDRSIZE) != 0
	  || memcmp (ipv6->dst, node_ipv6_addr.bits, IPv6_ADDRSIZE) != 0
	  || peek_u16_ntoh (tcp->src_port) != server_port
	  || peek_u16_ntoh (tcp->dst_port) != my_port)
	goto next_1;		/* Discard */

      if (((w = peek_u8 (tcp->flags)) & TCP_RST))
	return -1;

      if ((x = peek_u32_ntoh (tcp->ack)) != t_seq) /* SYN+ACK again */
	{
	  put_string ("Got SYN+ACK again\n");
	  t_seq -= send_size + 1;
	  goto syn_ack;
	}

      /* FIN-WAIT-2 or CLOSED */

      seq = peek_u32_ntoh (tcp->seq);
      len = peek_u16_ntoh (ipv6->payload_len) - (peek_u8 (tcp->hlen)>>2);
      offset = (peek_u8 (tcp->hlen)>>2) - sizeof (struct tcp);

      if (seq == r_seq)		/* New data */
	{
	  static int time = 0;
	  if ((++time & 0x1f) == 0x1f)
	    put_string (".");

	  if (receive_size < len)
	    {
	      put_string ("eth: Receive too big!\n");
	      tcp_reset (p, dst_eth_addr, server_ipv6_addr_p, server_port, my_port, t_seq);
	      return -1; /* XXX: upper layer have to call eth_maton to send RST...  */
	    }

	  if (len > 0)
	    {
	      memcpy (receive, tcp->data + offset, len);
	      receive += len;
	      receive_size -= len;
	      receive_count += len;
	      r_seq += len;
	      seq += len;
	      if (window_size > len)
		window_size -= len;
	      else
		window_size = 0;
	    }
	}

      if ((peek_u8 (tcp->flags) & TCP_FIN))
	{
	  if (seq == r_seq)
	    r_seq++; /* Fall through (next condition will met) */

	  if ((int)(r_seq - seq) > 0)
	    {			/* Already got FIN, restart timer */
	      count = 1;
	      retransmit_time = TIMEWAIT_TIMEOUT + count;
	    }
	}

      tcp_ack (p, t_seq, r_seq, window_size);
      continue;

    next_1:
      mdelay (1);
      if (count)
	count++;
    }

  return receive_count;
}

struct udp {
  u8 src_port[2];
  u8 dst_port[2];
  u8 len[2];
  u8 checksum[2];
  u8 data[0];
};

static void udp_header_build (struct eth_packet *p, u16 src_port, u16 dst_port, u16 len)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct udp *udp = (struct udp *)ipv6->data;
  u16 checksum;

  poke_u16_hton (udp->src_port, src_port);
  poke_u16_hton (udp->dst_port, dst_port);
  poke_u16_hton (udp->len, len);
  poke_u16_raw (udp->checksum, 0);
  checksum = tcpudp_checksum (ipv6, PROTO_UDP);
  if (checksum == 0)
    checksum = ~0;
  poke_u16_raw (udp->checksum, checksum);
}

#define DNS_SERVER 53
#define DNS_FLAG_QR_RESPONSE	0x8000
#define DNS_FLAG_RD		0x0100
#define DNS_QUERY_TYPE_AAAA	28
#define DNS_QUERY_CLASS_IN	1

struct dns {
  u8 ident[2];
  u8 flags[2];
  u8 num_questions[2];
  u8 num_answers[2];
  u8 num_authority[2];
  u8 num_additional[2];
  u8 data[0];
};

static void mk_dns_query (struct eth_packet *p,
			  const struct eth_addr *dst_eth_addr,
			  const struct ipv6_addr *dst_ipv6_addr_p,
			  const char *hostname,
			  u16 ident, u16 clientport)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct udp *udp = (struct udp *)ipv6->data;
  struct dns *dns = (struct dns *)udp->data;
  int len = strlen (hostname);
  int dns_size = sizeof (struct dns) + len + 2 + 2 + 2;
  int i;
  int n;

  p->length = ETH_HLEN + sizeof (struct ipv6) + sizeof (struct udp) + dns_size;
  ether_header_build (p, dst_eth_addr);
  ipv6_header_build (p, PROTO_UDP, sizeof (struct udp) + dns_size, HOP_LIMIT_DEFAULT, 
		     &node_ipv6_addr, dst_ipv6_addr_p);
  poke_u16_hton (dns->ident, ident);
  poke_u16_hton (dns->flags, DNS_FLAG_RD);
  poke_u16_hton (dns->num_questions, 1);
  poke_u16_hton (dns->num_answers, 0);
  poke_u16_hton (dns->num_authority, 0);
  poke_u16_hton (dns->num_additional, 0);
  memcpy (dns->data+1, hostname, len);
  poke_u8 (dns->data, '.');	/* Sentinel */
  poke_u8 (dns->data+1+len, 0);
  n=0;
  for (i=len; i>=0; i--)
    {
      if (peek_u8 (&dns->data[i]) == '.')
	{
	  poke_u8 (&dns->data[i], n);
	  n = 0;
	}
      else
	n++;
    }
  poke_u16_hton (dns->data+1+len+1, DNS_QUERY_TYPE_AAAA);
  poke_u16_hton (dns->data+1+len+1+2, DNS_QUERY_CLASS_IN);
  udp_header_build (p, clientport, DNS_SERVER, dns_size + sizeof (struct udp));
}

#define DNS_IDENT 0x1234
#define DNS_CLIENT_PORT 16337	/* arbitral */

/* Only support IN A query */
/* Rely on recursive query supported by the server */
int dns_resolv6 (struct eth_packet *p,
		 const struct eth_addr *dst_eth_addr,
		 const struct ipv6_addr *server_ipv6_addr_p,
		 const char *hostname, struct ipv6_addr *target_ipv6_addr_p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct icmpv6 *icmpv6 = (struct icmpv6 *)ipv6->data;
  struct udp *udp = (struct udp *)ipv6->data;
  struct dns *dns = (struct dns *)udp->data;
  int n;

  for (n = 0; n < 10; n++)
    {
      int i, r;

      mk_dns_query (p, dst_eth_addr, server_ipv6_addr_p, hostname, DNS_IDENT, DNS_CLIENT_PORT);
      transmit_prepare (p);

#define DNS_TIMEOUT 1000
      for (i = 0; i < DNS_TIMEOUT; i++)
	{
	  while ((r = ipv6_maton (p)) == 1);

	  if (r == 0)
	    {			/* Got packet */
	      if (peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
		continue;

	      if (p->length < (ETH_HLEN + sizeof (struct ipv6)
			       + sizeof (struct udp) + sizeof (struct dns)))
		continue;

	      if (peek_u8 (ipv6->next_header) == PROTO_ICMPv6)
		{			/* Error? */
		  if (peek_u8 (icmpv6->type) != ICMP_DESTINATION_UNREACHABLE)
		    continue;

		  /* ICMP_REDIRECT? ICMP_PARAMETER_PROBLEM? */
		  /* XXX: Should check if this is for this UDP connection */
		  return -1;
		}

	      if (peek_u8 (ipv6->next_header) != PROTO_UDP
		  || p->length < ETH_HLEN + sizeof (struct ipv6) + peek_u16_ntoh (ipv6->payload_len)
		  || peek_u16_ntoh (ipv6->payload_len) < sizeof (struct udp) + sizeof (struct dns)
		  || tcpudp_checksum (ipv6, PROTO_UDP) != 0
		  || memcmp (ipv6->src, server_ipv6_addr_p->bits, IPv6_ADDRSIZE) != 0
		  || memcmp (ipv6->dst, node_ipv6_addr.bits, IPv6_ADDRSIZE) != 0
		  || peek_u16_ntoh (udp->src_port) != DNS_SERVER
		  || peek_u16_ntoh (udp->dst_port) != DNS_CLIENT_PORT)
		continue;

	      if (peek_u16_ntoh (dns->ident) != DNS_IDENT)
		continue;

	      if ((peek_u16_ntoh (dns->flags) & DNS_FLAG_QR_RESPONSE) == 0)
		continue;

	      if (peek_u16_ntoh (dns->num_answers) == 0)
		return -1;	/* Failed */

	      /* OK, we got the answer. */
	      {
		u8 u, *q;

		/* Skip Query data */
		for (q = dns->data; q < p->data + p->length && (u = *q) != 0; )
		  if ((u & 0xc0) == 0xc0) /* Pointer */
		    {
		      q++;
		      break;
		    }
		  else
		    q += u + 1;
		q += 5;

		while (q < p->data + p->length && (u = *q) != 0)
		  if ((u & 0xc0) == 0xc0) /* Pointer */
		    {
		      q++;
		      break;
		    }
		  else
		    q += u + 1;
		q += 9;
		if (peek_u16_ntoh (q) != IPv6_ADDRSIZE)
		  return -1;
		q += 2;
		memcpy (target_ipv6_addr_p->bits, q, IPv6_ADDRSIZE);

		return 1;
	      }
	    }

	  mdelay (1);
	}
    }

  return -1;
}

struct dhcpv6 {
  u8 msg_type[1];
  u8 transaction_id[3];
  u8 options[0];
};

#define DHCPV6_SERVER 547
#define DHCPV6_CLIENT 546

#define DHCP_MSG_REPLY			7
#define DHCP_MSG_INFORMATION_REQUEST	11

struct dhcpv6_options {
  u8 code[2];
  u8 len[2];
  u8 data[0];
};

#define DHCP_OPTION_CLIENTID		1
#define DHCP_OPTION_SERVERID		2
#define DHCP_OPTION_ORO			6
#define DHCP_OPTION_STATUS_CODE		13
#define DHCP_OPTION_DNS_SERVERS		23
#define DHCP_OPTION_DOMAIN_LIST		24

const u8 dhcpv6_info_request_with_option_oro[] =
  {
    DHCP_MSG_INFORMATION_REQUEST, 
    0x00, 0x34, 0x56,		/* Transaction ID */
    0x00, DHCP_OPTION_CLIENTID, 
    0, 10, 
    0x00, 0x03, /* DDID-LL */
    0x00, 0x01, /* ARPHRD_ETHER */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* Ether address */
    0x00, DHCP_OPTION_ORO, 
    0, 4, 
    0x00, DHCP_OPTION_DNS_SERVERS, 
    0x00, DHCP_OPTION_DOMAIN_LIST, 
  };

struct dhcp_option_status_code {
  u8 code[2];
  u8 len[2];
  u8 status_code[1];
  u8 message[0];
};

#define DHCP_STATUS_SUCCESS		0

static void mk_dhcp6 (struct eth_packet *p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct udp *udp = (struct udp *)ipv6->data;
  u8 all_dhcp_servers[] = { 0x33, 0x33, 0x00, 0x01, 0x00, 0x02 };
  struct ipv6_addr srv = {{ 0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
			    0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02 }};

  p->length = (ETH_HLEN + sizeof (struct ipv6) + sizeof (struct udp)
	       + sizeof (dhcpv6_info_request_with_option_oro));
  ether_header_build (p, (struct eth_addr *)all_dhcp_servers);
  ipv6_header_build (p, PROTO_UDP, (sizeof (struct udp)
				    + sizeof (dhcpv6_info_request_with_option_oro)),
		     HOP_LIMIT_DEFAULT, &node_ipv6_addr, &srv);
  memcpy (udp->data, dhcpv6_info_request_with_option_oro, 
	  sizeof (dhcpv6_info_request_with_option_oro));
  memcpy (&udp->data[12], node_eth_addr.bits, ETH_ALEN);
  udp_header_build (p, DHCPV6_CLIENT, DHCPV6_SERVER,
		    sizeof (struct udp) + sizeof (dhcpv6_info_request_with_option_oro));
}

int dhcp6 (struct eth_packet *p, struct ipv6_addr *dns_ipv6_addr_p)
{
  struct eth_frame *e = (struct eth_frame *)p->data;
  struct ipv6 *ipv6 = (struct ipv6 *)e->data;
  struct udp *udp = (struct udp *)ipv6->data;
  struct dhcpv6 *dhcpv6 = (struct dhcpv6 *)udp->data;
  int r;
  int n;
  int len;
  u8 *q;
  int i;
  int found = 0;
  int l;

  for (n = 0; n < 3; n++)
    {
      mk_dhcp6 (p);
      transmit_prepare (p);

#define DHCP_TIMEOUT 1000
      for (i = 0; i < DHCP_TIMEOUT; i++)
	{
	  while ((r = ipv6_maton (p)) == 1);

	  if (r == 0)
	    {
	      if (peek_u16_ntoh (e->type) != ETHERTYPE_IPv6)
		continue;

	      if (p->length < ETH_HLEN + sizeof (struct ipv6) + sizeof (struct udp))
		continue;

	      len = peek_u16_ntoh (ipv6->payload_len) - sizeof (struct udp);
	      if (len < sizeof (struct dhcpv6)
		  || p->length < ETH_HLEN + sizeof (struct ipv6) + sizeof (struct udp) + len)
		continue;

	      if (peek_u8 (ipv6->next_header) != PROTO_UDP)
		continue;

	      if (peek_u16_raw (udp->checksum) != 0
		  && tcpudp_checksum (ipv6, PROTO_UDP) != 0)
		continue;

	      if (peek_u16_ntoh (udp->src_port) != DHCPV6_SERVER
		  || peek_u16_ntoh (udp->dst_port) != DHCPV6_CLIENT)
		continue;

	      if (peek_u8 (dhcpv6->msg_type) != DHCP_MSG_REPLY)
		continue;

	      if (memcmp (dhcpv6->transaction_id,
			  &dhcpv6_info_request_with_option_oro[1], 3) != 0)
		continue;

	      goto got_packet;
	    }

	  mdelay (1);
	}
    }

  return -1;

 got_packet:
  /* status code */
  /* DNS server */
  /* domain name list */
  /* possibly, server id */

  len -= sizeof (struct dhcpv6);
  for (q = dhcpv6->options; q < dhcpv6->options + len; q += l)
    {
      u16 code = peek_u16_ntoh (q);

      q += 2;
      l = peek_u16_ntoh (q);
      q += 2;

      switch (code)
	{
	case DHCP_OPTION_SERVERID:
	  break;

	case DHCP_OPTION_STATUS_CODE:
	  if (peek_u8 (q) != DHCP_STATUS_SUCCESS)
	    {
	      q[l] = 0;
	      put_string (q+1);
	      return -1;
	    }
	  break;

	case DHCP_OPTION_DNS_SERVERS:
	  if (l >= IPv6_ADDRSIZE) /* Use first one. */
	    {
	      memcpy (dns_ipv6_addr_p->bits, q, IPv6_ADDRSIZE);
	      found = 1;
	    }
	  break;

	case DHCP_OPTION_DOMAIN_LIST:
	  break;

	default:
	  break;
	}
    }

  if (found)
    return 1;

  return -1;
}
